import logging
import os
import shlex
import signal
import subprocess
import sys
from typing import Any, Dict, List, Optional

import wandb
from wandb.sdk.launch.builder.abstract import AbstractBuilder
from wandb.sdk.launch.environment.abstract import AbstractEnvironment

from .._project_spec import LaunchProject
from ..builder.build import get_env_vars_dict
from ..utils import (
    LOG_PREFIX,
    PROJECT_SYNCHRONOUS,
    _is_wandb_dev_uri,
    _is_wandb_local_uri,
    docker_image_exists,
    pull_docker_image,
    sanitize_wandb_api_key,
)
from .abstract import AbstractRun, AbstractRunner, Status

_logger = logging.getLogger(__name__)


class LocalSubmittedRun(AbstractRun):
    """Instance of ``AbstractRun`` corresponding to a subprocess launched to run an entry point command locally."""

    def __init__(self, command_proc: "subprocess.Popen[bytes]") -> None:
        super().__init__()
        self.command_proc = command_proc

    @property
    def id(self) -> str:
        return str(self.command_proc.pid)

    def wait(self) -> bool:
        return self.command_proc.wait() == 0

    def cancel(self) -> None:
        # Interrupt child process if it hasn't already exited
        if self.command_proc.poll() is None:
            # Kill the the process tree rooted at the child if it's the leader of its own process
            # group, otherwise just kill the child
            try:
                if self.command_proc.pid == os.getpgid(self.command_proc.pid):
                    os.killpg(self.command_proc.pid, signal.SIGTERM)
                else:
                    self.command_proc.terminate()
            except OSError:
                # The child process may have exited before we attempted to terminate it, so we
                # ignore OSErrors raised during child process termination
                _msg = f"{LOG_PREFIX}Failed to terminate child process PID {self.command_proc.pid}"
                _logger.debug(_msg)
            self.command_proc.wait()

    def get_status(self) -> Status:
        exit_code = self.command_proc.poll()
        if exit_code is None:
            return Status("running")
        if exit_code == 0:
            return Status("finished")
        return Status("failed")


class LocalContainerRunner(AbstractRunner):
    """Runner class, uses a project to create a LocallySubmittedRun."""

    def __init__(
        self,
        api: wandb.apis.internal.Api,
        backend_config: Dict[str, Any],
        environment: AbstractEnvironment,
    ) -> None:
        super().__init__(api, backend_config)
        self.environment = environment

    def run(
        self,
        launch_project: LaunchProject,
        builder: Optional[AbstractBuilder],
    ) -> Optional[AbstractRun]:
        synchronous: bool = self.backend_config[PROJECT_SYNCHRONOUS]
        docker_args: Dict[str, Any] = launch_project.resource_args.get(
            "local-container", {}
        )

        if _is_wandb_local_uri(self._api.settings("base_url")):
            if sys.platform == "win32":
                docker_args["net"] = "host"
            else:
                docker_args["network"] = "host"
            if sys.platform == "linux" or sys.platform == "linux2":
                docker_args["add-host"] = "host.docker.internal:host-gateway"

        entry_point = launch_project.get_single_entry_point()
        env_vars = get_env_vars_dict(launch_project, self._api)

        # When running against local port, need to swap to local docker host
        if (
            _is_wandb_local_uri(self._api.settings("base_url"))
            and sys.platform == "darwin"
        ):
            _, _, port = self._api.settings("base_url").split(":")
            env_vars["WANDB_BASE_URL"] = f"http://host.docker.internal:{port}"
        elif _is_wandb_dev_uri(self._api.settings("base_url")):
            env_vars["WANDB_BASE_URL"] = "http://host.docker.internal:9002"

        if launch_project.docker_image:
            # user has provided their own docker image
            image_uri = launch_project.image_name
            if not docker_image_exists(image_uri):
                pull_docker_image(image_uri)
            entry_cmd = []
            if entry_point is not None:
                entry_cmd = entry_point.command
            command_str = " ".join(
                get_docker_command(
                    image_uri,
                    env_vars,
                    entry_cmd=entry_cmd,
                    docker_args=docker_args,
                    additional_args=launch_project.override_args,
                )
            ).strip()
        else:
            assert entry_point is not None
            _logger.info("Building docker image...")
            assert builder is not None
            image_uri = builder.build_image(
                launch_project,
                entry_point,
            )
            _logger.info(f"Docker image built with uri {image_uri}")
            # entry_cmd and additional_args are empty here because
            # if launch built the container they've been accounted
            # in the dockerfile and env vars respectively
            command_str = " ".join(
                get_docker_command(
                    image_uri,
                    env_vars,
                    docker_args=docker_args,
                )
            ).strip()

        sanitized_cmd_str = sanitize_wandb_api_key(command_str)
        _msg = f"{LOG_PREFIX}Launching run in docker with command: {sanitized_cmd_str}"
        wandb.termlog(_msg)
        run = _run_entry_point(command_str, launch_project.project_dir)
        if synchronous:
            run.wait()
        return run


def _run_entry_point(command: str, work_dir: Optional[str]) -> AbstractRun:
    """Run an entry point command in a subprocess.

    Arguments:
        command: Entry point command to run
        work_dir: Working directory in which to run the command

    Returns:
        An instance of `LocalSubmittedRun`
    """
    if work_dir is None:
        work_dir = os.getcwd()
    env = os.environ.copy()
    if os.name == "nt":
        # we are running on windows
        process = subprocess.Popen(
            ["cmd", "/c", command], close_fds=True, cwd=work_dir, env=env
        )
    else:
        process = subprocess.Popen(
            ["bash", "-c", command],
            close_fds=True,
            cwd=work_dir,
            env=env,
        )

    return LocalSubmittedRun(process)


def get_docker_command(
    image: str,
    env_vars: Dict[str, str],
    entry_cmd: Optional[List[str]] = None,
    docker_args: Optional[Dict[str, Any]] = None,
    additional_args: Optional[List[str]] = None,
) -> List[str]:
    """Construct the docker command using the image and docker args.

    Arguments:
    image: a Docker image to be run
    env_vars: a dictionary of environment variables for the command
    entry_cmd: the entry point command to run
    docker_args: a dictionary of additional docker args for the command
    """
    docker_path = "docker"
    cmd: List[Any] = [docker_path, "run", "--rm"]

    # hacky handling of env vars, needs to be improved
    for env_key, env_value in env_vars.items():
        cmd += ["-e", f"{shlex.quote(env_key)}={shlex.quote(env_value)}"]

    if docker_args:
        for name, value in docker_args.items():
            if len(name) == 1:
                prefix = "-" + shlex.quote(name)
            else:
                prefix = "--" + shlex.quote(name)
            if isinstance(value, list):
                for v in value:
                    cmd += [prefix, shlex.quote(str(v))]
            elif isinstance(value, bool) and value:
                cmd += [prefix]
            else:
                cmd += [prefix, shlex.quote(str(value))]

    if entry_cmd:
        cmd += ["--entrypoint", entry_cmd[0]]
    cmd += [shlex.quote(image)]
    if entry_cmd and len(entry_cmd) > 1:
        cmd += entry_cmd[1:]
    if additional_args:
        cmd += additional_args
    return cmd


def join(split_command: List[str]) -> str:
    """Return a shell-escaped string from *split_command*."""
    return " ".join(shlex.quote(arg) for arg in split_command)



import json

from .visitor import Visitor, visit

__all__ = ['print_ast']


def print_ast(ast):
    return visit(ast, PrintingVisitor())


class PrintingVisitor(Visitor):
    __slots__ = ()

    def leave_Name(self, node, *args):
        return node.value

    def leave_Variable(self, node, *args):
        return '$' + node.name

    def leave_Document(self, node, *args):
        return join(node.definitions, '\n\n') + '\n'

    def leave_OperationDefinition(self, node, *args):
        name = node.name
        selection_set = node.selection_set
        op = node.operation
        var_defs = wrap('(', join(node.variable_definitions, ', '), ')')
        directives = join(node.directives, ' ')

        if not name and not directives and not var_defs and op == 'query':
            return selection_set

        return join([op, join([name, var_defs]), directives, selection_set], ' ')

    def leave_VariableDefinition(self, node, *args):
        return node.variable + ': ' + node.type + wrap(' = ', node.default_value)

    def leave_SelectionSet(self, node, *args):
        return block(node.selections)

    def leave_Field(self, node, *args):
        return join([
            wrap('', node.alias, ': ') + node.name + wrap('(', join(node.arguments, ', '), ')'),
            join(node.directives, ' '),
            node.selection_set
        ], ' ')

    def leave_Argument(self, node, *args):
        return node.name + ': ' + node.value

    # Fragments

    def leave_FragmentSpread(self, node, *args):
        return '...' + node.name + wrap(' ', join(node.directives, ' '))

    def leave_InlineFragment(self, node, *args):
        return join([
            '...',
            wrap('on ', node.type_condition),
            join(node.directives, ''),
            node.selection_set
        ], ' ')

    def leave_FragmentDefinition(self, node, *args):
        return ('fragment {} on {} '.format(node.name, node.type_condition) +
                wrap('', join(node.directives, ' '), ' ') +
                node.selection_set)

    # Value

    def leave_IntValue(self, node, *args):
        return node.value

    def leave_FloatValue(self, node, *args):
        return node.value

    def leave_StringValue(self, node, *args):
        return json.dumps(node.value)

    def leave_BooleanValue(self, node, *args):
        return json.dumps(node.value)

    def leave_EnumValue(self, node, *args):
        return node.value

    def leave_ListValue(self, node, *args):
        return '[' + join(node.values, ', ') + ']'

    def leave_ObjectValue(self, node, *args):
        return '{' + join(node.fields, ', ') + '}'

    def leave_ObjectField(self, node, *args):
        return node.name + ': ' + node.value

    # Directive

    def leave_Directive(self, node, *args):
        return '@' + node.name + wrap('(', join(node.arguments, ', '), ')')

    # Type

    def leave_NamedType(self, node, *args):
        return node.name

    def leave_ListType(self, node, *args):
        return '[' + node.type + ']'

    def leave_NonNullType(self, node, *args):
        return node.type + '!'

    # Type Definitions:

    def leave_SchemaDefinition(self, node, *args):
        return join([
            'schema',
            join(node.directives, ' '),
            block(node.operation_types),
            ], ' ')

    def leave_OperationTypeDefinition(self, node, *args):
        return '{}: {}'.format(node.operation, node.type)

    def leave_ScalarTypeDefinition(self, node, *args):
        return 'scalar ' + node.name + wrap(' ', join(node.directives, ' '))

    def leave_ObjectTypeDefinition(self, node, *args):
        return join([
            'type',
            node.name,
            wrap('implements ', join(node.interfaces, ', ')),
            join(node.directives, ' '),
            block(node.fields)
        ], ' ')

    def leave_FieldDefinition(self, node, *args):
        return (
            node.name +
            wrap('(', join(node.arguments, ', '), ')') +
            ': ' +
            node.type +
            wrap(' ', join(node.directives, ' '))
        )

    def leave_InputValueDefinition(self, node, *args):
        return node.name + ': ' + node.type + wrap(' = ', node.default_value) + wrap(' ', join(node.directives, ' '))

    def leave_InterfaceTypeDefinition(self, node, *args):
        return 'interface ' + node.name + wrap(' ', join(node.directives, ' ')) + ' ' + block(node.fields)

    def leave_UnionTypeDefinition(self, node, *args):
        return 'union ' + node.name + wrap(' ', join(node.directives, ' ')) + ' = ' + join(node.types, ' | ')

    def leave_EnumTypeDefinition(self, node, *args):
        return 'enum ' + node.name + wrap(' ', join(node.directives, ' ')) + ' ' + block(node.values)

    def leave_EnumValueDefinition(self, node, *args):
        return node.name + wrap(' ', join(node.directives, ' '))

    def leave_InputObjectTypeDefinition(self, node, *args):
        return 'input ' + node.name + wrap(' ', join(node.directives, ' ')) + ' ' + block(node.fields)

    def leave_TypeExtensionDefinition(self, node, *args):
        return 'extend ' + node.definition

    def leave_DirectiveDefinition(self, node, *args):
        return 'directive @{}{} on {}'.format(node.name, wrap(
            '(', join(node.arguments, ', '), ')'), ' | '.join(node.locations))


def join(maybe_list, separator=''):
    if maybe_list:
        return separator.join(filter(None, maybe_list))
    return ''


def block(_list):
    '''Given a list, print each item on its own line, wrapped in an indented "{ }" block.'''
    if _list:
        return indent('{\n' + join(_list, '\n')) + '\n}'
    return '{}'


def wrap(start, maybe_str, end=''):
    if maybe_str:
        return start + maybe_str + end
    return ''


def indent(maybe_str):
    if maybe_str:
        return maybe_str.replace('\n', '\n  ')
    return maybe_str



# This tests reported issues in the Promise package
from concurrent.futures import ThreadPoolExecutor
from promise import Promise
import time
import weakref
import gc

executor = ThreadPoolExecutor(max_workers=40000)


def test_issue_11():
    # https://github.com/syrusakbary/promise/issues/11
    def test(x):
        def my(resolve, reject):
            if x > 0:
                resolve(x)
            else:
                reject(Exception(x))

        return Promise(my)

    promise_resolved = test(42).then(lambda x: x)
    assert promise_resolved.get() == 42

    promise_rejected = test(-42).then(lambda x: x, lambda e: str(e))
    assert promise_rejected.get() == "-42"


def identity(x, wait):
    if wait:
        time.sleep(wait)
    return x


def promise_with_wait(x, wait):
    return Promise.resolve(identity(x, wait))


def test_issue_9():
    no_wait = Promise.all(
        [promise_with_wait(x1, None).then(lambda y: x1 * y) for x1 in (0, 1, 2, 3)]
    ).get()
    wait_a_bit = Promise.all(
        [promise_with_wait(x2, 0.05).then(lambda y: x2 * y) for x2 in (0, 1, 2, 3)]
    ).get()
    wait_longer = Promise.all(
        [promise_with_wait(x3, 0.1).then(lambda y: x3 * y) for x3 in (0, 1, 2, 3)]
    ).get()

    assert no_wait == wait_a_bit
    assert no_wait == wait_longer


@Promise.safe
def test_issue_9_safe():
    no_wait = Promise.all(
        [promise_with_wait(x1, None).then(lambda y: x1 * y) for x1 in (0, 1, 2, 3)]
    ).get()
    wait_a_bit = Promise.all(
        [promise_with_wait(x2, 0.05).then(lambda y: x2 * y) for x2 in (0, 1, 2, 3)]
    ).get()
    wait_longer = Promise.all(
        [promise_with_wait(x3, 0.1).then(lambda y: x3 * y) for x3 in (0, 1, 2, 3)]
    ).get()

    assert no_wait == [0, 3, 6, 9]
    assert no_wait == wait_a_bit
    assert no_wait == wait_longer


def test_issue_26():
    context = {"success": False}
    promise1 = Promise(
        lambda resolve, reject: context.update({"promise1_reject": reject})
    )
    promise1.then(lambda x: None)
    promise1.then(lambda x: None)
    context["promise1_reject"](RuntimeError("Ooops!"))

    promise2 = Promise(
        lambda resolve, reject: context.update({"promise2_resolve": resolve})
    )
    promise3 = promise2.then(lambda x: context.update({"success": True}))
    context["promise2_resolve"](None)

    # We wait so it works in asynchronous envs
    promise3._wait(timeout=.1)
    assert context["success"]


# def promise_in_executor(x, wait):
#     return Promise.promisify(executor.submit(identity, x, wait))


# @Promise.safe
# def test_issue_9_extra():
#     no_wait = Promise.all([promise_in_executor(x1, None).then(lambda y: x1*y) for x1 in (0,1,2,3)]).get()
#     wait_a_bit = Promise.all([promise_in_executor(x2, 0.1).then(lambda y: x2*y) for x2 in (0,1,2,3)]).get()
#     wait_longer = Promise.all([promise_in_executor(x3, 0.5).then(lambda y: x3*y) for x3 in (0,1,2,3)]).get()

#     assert no_wait == [0, 3, 6, 9]
#     assert no_wait == wait_a_bit
#     assert no_wait == wait_longer


def test_issue_33():
    def do(x):
        v = Promise.resolve("ok").then(lambda x: x).get()
        return v

    p = Promise.resolve(None).then(do)
    assert p.get() == "ok"


def test_issue_75():
    def function_with_local_type():
        class A:
            pass

        a = A()
        assert a == Promise.resolve(a).get()

        return weakref.ref(A)

    weak_reference = function_with_local_type()

    # The local type 'A' from the function is still kept alive by reference cycles.
    gc.collect()

    # Now the local type should have been garbage collected,
    # such that the weak reference should be invalid.
    assert not weak_reference()



from ..execution.values import get_argument_values
from ..language import ast
from ..pyutils.ordereddict import OrderedDict
from ..type import (GraphQLArgument, GraphQLBoolean,
                    GraphQLDeprecatedDirective, GraphQLDirective,
                    GraphQLEnumType, GraphQLEnumValue, GraphQLField,
                    GraphQLFloat, GraphQLID, GraphQLIncludeDirective,
                    GraphQLInputObjectField, GraphQLInputObjectType,
                    GraphQLInt, GraphQLInterfaceType, GraphQLList,
                    GraphQLNonNull, GraphQLObjectType, GraphQLScalarType,
                    GraphQLSchema, GraphQLSkipDirective, GraphQLString,
                    GraphQLUnionType)
from ..type.introspection import (__Directive, __DirectiveLocation,
                                  __EnumValue, __Field, __InputValue, __Schema,
                                  __Type, __TypeKind)
from ..utils.value_from_ast import value_from_ast


def _build_wrapped_type(inner_type, input_type_ast):
    if isinstance(input_type_ast, ast.ListType):
        return GraphQLList(_build_wrapped_type(inner_type, input_type_ast.type))

    if isinstance(input_type_ast, ast.NonNullType):
        return GraphQLNonNull(_build_wrapped_type(inner_type, input_type_ast.type))

    return inner_type


def _get_inner_type_name(type_ast):
    if isinstance(type_ast, (ast.ListType, ast.NonNullType)):
        return _get_inner_type_name(type_ast.type)

    return type_ast.name.value


def _get_named_type_ast(type_ast):
    named_type = type_ast
    while isinstance(named_type, (ast.ListType, ast.NonNullType)):
        named_type = named_type.type

    return named_type


def _false(*_):
    return False


def _none(*_):
    return None


def build_ast_schema(document):
    assert isinstance(document, ast.Document), 'must pass in Document ast.'

    schema_def = None

    type_asts = (
        ast.ScalarTypeDefinition,
        ast.ObjectTypeDefinition,
        ast.InterfaceTypeDefinition,
        ast.EnumTypeDefinition,
        ast.UnionTypeDefinition,
        ast.InputObjectTypeDefinition,
    )

    type_defs = []
    directive_defs = []

    for d in document.definitions:
        if isinstance(d, ast.SchemaDefinition):
            if schema_def:
                raise Exception('Must provide only one schema definition.')
            schema_def = d
        if isinstance(d, type_asts):
            type_defs.append(d)
        elif isinstance(d, ast.DirectiveDefinition):
            directive_defs.append(d)

    if not schema_def:
        raise Exception('Must provide a schema definition.')

    query_type_name = None
    mutation_type_name = None
    subscription_type_name = None

    for operation_type in schema_def.operation_types:
        type_name = operation_type.type.name.value
        if operation_type.operation == 'query':
            if query_type_name:
                raise Exception('Must provide only one query type in schema.')
            query_type_name = type_name
        elif operation_type.operation == 'mutation':
            if mutation_type_name:
                raise Exception('Must provide only one mutation type in schema.')
            mutation_type_name = type_name
        elif operation_type.operation == 'subscription':
            if subscription_type_name:
                raise Exception('Must provide only one subscription type in schema.')
            subscription_type_name = type_name

    if not query_type_name:
        raise Exception('Must provide schema definition with query type.')

    ast_map = {d.name.value: d for d in type_defs}

    if query_type_name not in ast_map:
        raise Exception('Specified query type "{}" not found in document.'.format(query_type_name))

    if mutation_type_name and mutation_type_name not in ast_map:
        raise Exception('Specified mutation type "{}" not found in document.'.format(mutation_type_name))

    if subscription_type_name and subscription_type_name not in ast_map:
        raise Exception('Specified subscription type "{}" not found in document.'.format(subscription_type_name))

    inner_type_map = OrderedDict([
        ('String', GraphQLString),
        ('Int', GraphQLInt),
        ('Float', GraphQLFloat),
        ('Boolean', GraphQLBoolean),
        ('ID', GraphQLID),
        ('__Schema', __Schema),
        ('__Directive', __Directive),
        ('__DirectiveLocation', __DirectiveLocation),
        ('__Type', __Type),
        ('__Field', __Field),
        ('__InputValue', __InputValue),
        ('__EnumValue', __EnumValue),
        ('__TypeKind', __TypeKind),
    ])

    def get_directive(directive_ast):
        return GraphQLDirective(
            name=directive_ast.name.value,
            locations=[node.value for node in directive_ast.locations],
            args=make_input_values(directive_ast.arguments, GraphQLArgument),
        )

    def get_object_type(type_ast):
        type = type_def_named(type_ast.name.value)
        assert isinstance(type, GraphQLObjectType), 'AST must provide object type'
        return type

    def produce_type_def(type_ast):
        type_name = _get_named_type_ast(type_ast).name.value
        type_def = type_def_named(type_name)
        return _build_wrapped_type(type_def, type_ast)

    def type_def_named(type_name):
        if type_name in inner_type_map:
            return inner_type_map[type_name]

        if type_name not in ast_map:
            raise Exception('Type "{}" not found in document'.format(type_name))

        inner_type_def = make_schema_def(ast_map[type_name])
        if not inner_type_def:
            raise Exception('Nothing constructed for "{}".'.format(type_name))

        inner_type_map[type_name] = inner_type_def
        return inner_type_def

    def make_schema_def(definition):
        if not definition:
            raise Exception('def must be defined.')

        handler = _schema_def_handlers.get(type(definition))
        if not handler:
            raise Exception('Type kind "{}" not supported.'.format(type(definition).__name__))

        return handler(definition)

    def make_type_def(definition):
        return GraphQLObjectType(
            name=definition.name.value,
            fields=lambda: make_field_def_map(definition),
            interfaces=make_implemented_interfaces(definition)
        )

    def make_field_def_map(definition):
        return OrderedDict(
            (f.name.value, GraphQLField(
                type=produce_type_def(f.type),
                args=make_input_values(f.arguments, GraphQLArgument),
                deprecation_reason=get_deprecation_reason(f.directives),
            ))
            for f in definition.fields
        )

    def make_implemented_interfaces(definition):
        return [produce_type_def(i) for i in definition.interfaces]

    def make_input_values(values, cls):
        return OrderedDict(
            (value.name.value, cls(
                type=produce_type_def(value.type),
                default_value=value_from_ast(value.default_value, produce_type_def(value.type))
            ))
            for value in values
        )

    def make_interface_def(definition):
        return GraphQLInterfaceType(
            name=definition.name.value,
            resolve_type=_none,
            fields=lambda: make_field_def_map(definition)
        )

    def make_enum_def(definition):
        values = OrderedDict((v.name.value, GraphQLEnumValue(deprecation_reason=get_deprecation_reason(v.directives)))
                             for v in definition.values)
        return GraphQLEnumType(
            name=definition.name.value,
            values=values
        )

    def make_union_def(definition):
        return GraphQLUnionType(
            name=definition.name.value,
            resolve_type=_none,
            types=[produce_type_def(t) for t in definition.types]
        )

    def make_scalar_def(definition):
        return GraphQLScalarType(
            name=definition.name.value,
            serialize=_none,
            # Validation calls the parse functions to determine if a literal value is correct.
            # Returning none, however would cause the scalar to fail validation. Returning false,
            # will cause them to pass.
            parse_literal=_false,
            parse_value=_false
        )

    def make_input_object_def(definition):
        return GraphQLInputObjectType(
            name=definition.name.value,
            fields=make_input_values(definition.fields, GraphQLInputObjectField)
        )

    _schema_def_handlers = {
        ast.ObjectTypeDefinition: make_type_def,
        ast.InterfaceTypeDefinition: make_interface_def,
        ast.EnumTypeDefinition: make_enum_def,
        ast.UnionTypeDefinition: make_union_def,
        ast.ScalarTypeDefinition: make_scalar_def,
        ast.InputObjectTypeDefinition: make_input_object_def
    }
    types = [type_def_named(definition.name.value) for definition in type_defs]
    directives = [get_directive(d) for d in directive_defs]

    # If specified directive were not explicitly declared, add them.
    find_skip_directive = (directive.name for directive in directives if directive.name == 'skip')
    find_include_directive = (directive.name for directive in directives if directive.name == 'include')
    find_deprecated_directive = (directive.name for directive in directives if directive.name == 'deprecated')

    if not next(find_skip_directive, None):
        directives.append(GraphQLSkipDirective)

    if not next(find_include_directive, None):
        directives.append(GraphQLIncludeDirective)

    if not next(find_deprecated_directive, None):
        directives.append(GraphQLDeprecatedDirective)

    schema_kwargs = {'query': get_object_type(ast_map[query_type_name])}

    if mutation_type_name:
        schema_kwargs['mutation'] = get_object_type(ast_map[mutation_type_name])

    if subscription_type_name:
        schema_kwargs['subscription'] = get_object_type(ast_map[subscription_type_name])

    if directive_defs:
        schema_kwargs['directives'] = directives

    if types:
        schema_kwargs['types'] = types

    return GraphQLSchema(**schema_kwargs)


def get_deprecation_reason(directives):
    deprecated_ast = next((directive for directive in directives
                          if directive.name.value == GraphQLDeprecatedDirective.name),
                          None)

    if deprecated_ast:
        args = get_argument_values(GraphQLDeprecatedDirective.args, deprecated_ast.arguments)
        return args['reason']
    else:
        return None



from wandb_gql import gql

VIEW_REPORT = gql(
    """
    query SpecificReport($reportId: ID!) {
        view(id: $reportId) {
        id
        type
        name
        displayName
        description
        project {
            id
            name
            entityName
        }
        createdAt
        updatedAt
        spec
        previewUrl
        user {
            name
            username
            userInfo
        }
        }
    }
    """
)
UPSERT_VIEW = gql(
    """
    mutation upsertView(
        $id: ID
        $entityName: String
        $projectName: String
        $type: String
        $name: String
        $displayName: String
        $description: String
        $spec: String!
    ) {
        upsertView(
        input: {
            id: $id
            entityName: $entityName
            projectName: $projectName
            name: $name
            displayName: $displayName
            description: $description
            type: $type
            createdUsing: WANDB_SDK
            spec: $spec
        }
        ) {
        view {
            id
            type
            name
            displayName
            description
            project {
                id
                name
            entityName
            }
            spec
            updatedAt
        }
        inserted
        }
    }
"""
)



"""grpc service.

Implement ServiceInterface for grpc transport.
"""

from typing import TYPE_CHECKING, Optional

import grpc

from wandb.proto import wandb_server_pb2 as spb
from wandb.proto import wandb_server_pb2_grpc as pbgrpc

from .service_base import ServiceInterface, _pbmap_apply_dict

if TYPE_CHECKING:
    from wandb.sdk.wandb_settings import Settings


class ServiceGrpcInterface(ServiceInterface):
    _stub: Optional[pbgrpc.InternalServiceStub]

    def __init__(self) -> None:
        self._stub = None

    def get_transport(self) -> str:
        return "grpc"

    def _svc_connect(self, port: int) -> None:
        channel = grpc.insecure_channel(f"localhost:{port}")
        stub = pbgrpc.InternalServiceStub(channel)
        self._stub = stub
        # TODO: make sure service is up

    def _get_stub(self) -> pbgrpc.InternalServiceStub:
        assert self._stub
        return self._stub

    def _svc_inform_init(self, settings: "Settings", run_id: str) -> None:
        inform_init = spb.ServerInformInitRequest()
        settings_dict = settings.make_static()
        _pbmap_apply_dict(inform_init._settings_map, settings_dict)
        inform_init._info.stream_id = run_id

        assert self._stub
        _ = self._stub.ServerInformInit(inform_init)

    def _svc_inform_start(self, settings: "Settings", run_id: str) -> None:
        inform_start = spb.ServerInformStartRequest()
        settings_dict = settings.make_static()
        _pbmap_apply_dict(inform_start._settings_map, settings_dict)
        inform_start._info.stream_id = run_id

        assert self._stub
        _ = self._stub.ServerInformStart(inform_start)

    def _svc_inform_finish(self, run_id: Optional[str] = None) -> None:
        assert run_id
        inform_finish = spb.ServerInformFinishRequest()
        inform_finish._info.stream_id = run_id

        assert self._stub
        _ = self._stub.ServerInformFinish(inform_finish)

    def _svc_inform_attach(self, attach_id: str) -> spb.ServerInformAttachResponse:
        assert self._stub

        inform_attach = spb.ServerInformAttachRequest()
        inform_attach._info.stream_id = attach_id
        return self._stub.ServerInformAttach(inform_attach)  # type: ignore

    def _svc_inform_teardown(self, exit_code: int) -> None:
        inform_teardown = spb.ServerInformTeardownRequest(exit_code=exit_code)

        assert self._stub
        _ = self._stub.ServerInformTeardown(inform_teardown)



from collections import OrderedDict, defaultdict
from collections.abc import Sequence
from functools import reduce

from ..utils.type_comparators import is_equal_type, is_type_sub_type_of
from .definition import (GraphQLArgument, GraphQLField,
                         GraphQLInputObjectField, GraphQLInputObjectType,
                         GraphQLInterfaceType, GraphQLList, GraphQLNonNull,
                         GraphQLObjectType, GraphQLUnionType, is_input_type,
                         is_output_type)


class GraphQLTypeMap(OrderedDict):

    def __init__(self, types):
        super(GraphQLTypeMap, self).__init__()
        self.update(reduce(self.reducer, types, OrderedDict()))
        self._possible_type_map = defaultdict(set)

        # Keep track of all implementations by interface name.
        self._implementations = {}
        for gql_type in self.values():
            if isinstance(gql_type, GraphQLObjectType):
                for interface in gql_type.interfaces:
                    self._implementations.setdefault(interface.name, []).append(gql_type)

        # Enforce correct interface implementations.
        for type in self.values():
            if isinstance(type, GraphQLObjectType):
                for interface in type.interfaces:
                    self.assert_object_implements_interface(self, type, interface)

    def get_possible_types(self, abstract_type):
        if isinstance(abstract_type, GraphQLUnionType):
            return abstract_type.types
        assert isinstance(abstract_type, GraphQLInterfaceType)
        return self._implementations.get(abstract_type.name, None)

    def is_possible_type(self, abstract_type, possible_type):
        possible_types = self.get_possible_types(abstract_type)
        assert isinstance(possible_types, Sequence), (
            'Could not find possible implementing types for ${} in ' +
            'schema. Check that schema.types is defined and is an array of' +
            'all possible types in the schema.'
            ).format(abstract_type)

        if not self._possible_type_map[abstract_type.name]:
            self._possible_type_map[abstract_type.name].update([p.name for p in possible_types])

        return possible_type.name in self._possible_type_map[abstract_type.name]

    @classmethod
    def reducer(cls, map, type):
        if not type:
            return map

        if isinstance(type, GraphQLList) or isinstance(type, GraphQLNonNull):
            return cls.reducer(map, type.of_type)

        if type.name in map:
            assert map[type.name] == type, (
                'Schema must contain unique named types but contains multiple types named "{}".'
            ).format(type.name)

            return map

        map[type.name] = type

        reduced_map = map

        if isinstance(type, (GraphQLUnionType)):
            for t in type.types:
                reduced_map = cls.reducer(reduced_map, t)

        if isinstance(type, GraphQLObjectType):
            for t in type.interfaces:
                reduced_map = cls.reducer(reduced_map, t)

        if isinstance(type, (GraphQLObjectType, GraphQLInterfaceType, GraphQLInputObjectType)):
            field_map = type.fields
            type_is_input = isinstance(type, GraphQLInputObjectType)
            for field_name, field in field_map.items():
                if type_is_input:
                    assert isinstance(field, GraphQLInputObjectField), (
                        '{}.{} must be an instance of GraphQLInputObjectField.'.format(type, field_name)
                    )
                    assert is_input_type(field.type), (
                        '{}.{} field type must be Input Type but got: {}.'.format(type, field_name, field.type)
                    )
                else:
                    assert isinstance(field, (GraphQLField, GraphQLField)), (
                        '{}.{} must be an instance of GraphQLField.'.format(type, field_name)
                    )
                    assert is_output_type(field.type), (
                        '{}.{} field type must be Output Type but got: {}.'.format(type, field_name, field.type)
                    )
                    for arg_name, arg in field.args.items():
                        assert isinstance(arg, (GraphQLArgument, GraphQLArgument)), (
                            '{}.{}({}:) argument must be an instance of GraphQLArgument.'.format(type, field_name, arg_name)
                        )
                        assert is_input_type(arg.type), (
                            '{}.{}({}:) argument type must be Input Type but got: {}.'.format(type, field_name, arg_name,
                                                                                              arg.type)
                        )
                        reduced_map = cls.reducer(reduced_map, arg.type)

                reduced_map = cls.reducer(reduced_map, getattr(field, 'type', None))

        return reduced_map

    @classmethod
    def assert_object_implements_interface(cls, schema, object, interface):
        object_field_map = object.fields
        interface_field_map = interface.fields

        for field_name, interface_field in interface_field_map.items():
            object_field = object_field_map.get(field_name)

            assert object_field, '"{}" expects field "{}" but "{}" does not provide it.'.format(
                interface, field_name, object
            )

            assert is_type_sub_type_of(schema, object_field.type, interface_field.type), (
                '{}.{} expects type "{}" but {}.{} provides type "{}".'
            ).format(interface, field_name, interface_field.type, object, field_name, object_field.type)

            for arg_name, interface_arg in interface_field.args.items():
                object_arg = object_field.args.get(arg_name)

                assert object_arg, (
                    '{}.{} expects argument "{}" but {}.{} does not provide it.'
                ).format(interface, field_name, arg_name, object, field_name)

                assert is_equal_type(interface_arg.type, object_arg.type), (
                    '{}.{}({}:) expects type "{}" but {}.{}({}:) provides type "{}".'
                ).format(interface, field_name, arg_name, interface_arg.type, object, field_name, arg_name, object_arg.type)

            for arg_name, object_arg in object_field.args.items():
                interface_arg = interface_field.args.get(arg_name)
                if not interface_arg:
                    assert not isinstance(object_arg.type, GraphQLNonNull), (
                        '{}.{}({}:) is of required type '
                        '"{}" but is not also provided by the '
                        'interface {}.{}.'
                    ).format(object, field_name, arg_name, object_arg.type, interface, field_name)



from typing import Any, Callable, Optional

import wandb


class PreInitObject:
    def __init__(self, name: str, destination: Optional[Any] = None) -> None:
        self._name = name

        if destination is not None:
            self.__doc__ = destination.__doc__

    def __getitem__(self, key: str) -> None:
        raise wandb.Error(f"You must call wandb.init() before {self._name}[{key!r}]")

    def __setitem__(self, key: str, value: Any) -> Any:
        raise wandb.Error(f"You must call wandb.init() before {self._name}[{key!r}]")

    def __setattr__(self, key: str, value: Any) -> Any:
        if not key.startswith("_"):
            raise wandb.Error(f"You must call wandb.init() before {self._name}.{key}")
        else:
            return object.__setattr__(self, key, value)

    def __getattr__(self, key: str) -> Any:
        if not key.startswith("_"):
            raise wandb.Error(f"You must call wandb.init() before {self._name}.{key}")
        else:
            raise AttributeError


def PreInitCallable(  # noqa: N802
    name: str, destination: Optional[Any] = None
) -> Callable:
    def preinit_wrapper(*args: Any, **kwargs: Any) -> Any:
        raise wandb.Error(f"You must call wandb.init() before {name}()")

    preinit_wrapper.__name__ = str(name)
    if destination:
        preinit_wrapper.__wrapped__ = destination  # type: ignore
        preinit_wrapper.__doc__ = destination.__doc__
    return preinit_wrapper



"""Implementation of the SageMakerRunner class."""
import logging
import time
from typing import Any, Dict, Optional, cast

if False:
    import boto3  # type: ignore

import wandb
from wandb.apis.internal import Api
from wandb.sdk.launch.builder.abstract import AbstractBuilder
from wandb.sdk.launch.environment.aws_environment import AwsEnvironment
from wandb.sdk.launch.utils import LaunchError

from .._project_spec import LaunchProject, get_entry_point_command
from ..builder.build import get_env_vars_dict
from ..utils import LOG_PREFIX, PROJECT_SYNCHRONOUS, to_camel_case
from .abstract import AbstractRun, AbstractRunner, Status

_logger = logging.getLogger(__name__)


class SagemakerSubmittedRun(AbstractRun):
    """Instance of ``AbstractRun`` corresponding to a subprocess launched to run an entry point command on aws sagemaker."""

    def __init__(self, training_job_name: str, client: "boto3.Client") -> None:
        super().__init__()
        self.client = client
        self.training_job_name = training_job_name
        self._status = Status("running")

    @property
    def id(self) -> str:
        return f"sagemaker-{self.training_job_name}"

    def wait(self) -> bool:
        while True:
            status_state = self.get_status().state
            wandb.termlog(
                f"{LOG_PREFIX}Training job {self.training_job_name} status: {status_state}"
            )
            if status_state in ["stopped", "failed", "finished"]:
                break
            time.sleep(5)
        return status_state == "finished"

    def cancel(self) -> None:
        # Interrupt child process if it hasn't already exited
        status = self.get_status()
        if status.state == "running":
            self.client.stop_training_job(TrainingJobName=self.training_job_name)
            self.wait()

    def get_status(self) -> Status:
        job_status = self.client.describe_training_job(
            TrainingJobName=self.training_job_name
        )["TrainingJobStatus"]
        if job_status == "Completed" or job_status == "Stopped":
            self._status = Status("finished")
        elif job_status == "Failed":
            self._status = Status("failed")
        elif job_status == "Stopping":
            self._status = Status("stopping")
        elif job_status == "InProgress":
            self._status = Status("running")
        return self._status


class SageMakerRunner(AbstractRunner):
    """Runner class, uses a project to create a SagemakerSubmittedRun."""

    def __init__(
        self, api: Api, backend_config: Dict[str, Any], environment: AwsEnvironment
    ) -> None:
        """Initialize the SagemakerRunner.

        Arguments:
            api (Api): The API instance.
            backend_config (Dict[str, Any]): The backend configuration.
            environment (AwsEnvironment): The AWS environment.

        Raises:
            LaunchError: If the runner cannot be initialized.
        """
        super().__init__(api, backend_config)
        self.environment = environment

    def run(
        self,
        launch_project: LaunchProject,
        builder: Optional[AbstractBuilder],
    ) -> Optional[AbstractRun]:
        """Run a project on Amazon Sagemaker.

        Arguments:
            launch_project (LaunchProject): The project to run.
            builder (AbstractBuilder): The builder to use.

        Returns:
            Optional[AbstractRun]: The run instance.

        Raises:
            LaunchError: If the launch is unsuccessful.
        """
        _logger.info("using AWSSagemakerRunner")

        given_sagemaker_args = launch_project.resource_args.get("sagemaker")
        if given_sagemaker_args is None:
            raise LaunchError(
                "No sagemaker args specified. Specify sagemaker args in resource_args"
            )

        default_output_path = self.backend_config.get("runner", {}).get(
            "s3_output_path"
        )
        if default_output_path is not None and not default_output_path.startswith(
            "s3://"
        ):
            default_output_path = f"s3://{default_output_path}"

        session = self.environment.get_session()
        client = session.client("sts")
        caller_id = client.get_caller_identity()
        account_id = caller_id["Account"]
        _logger.info(f"Using account ID {account_id}")
        role_arn = get_role_arn(given_sagemaker_args, self.backend_config, account_id)
        entry_point = launch_project.get_single_entry_point()

        # Create a sagemaker client to launch the job.
        sagemaker_client = session.client("sagemaker")

        # if the user provided the image they want to use, use that, but warn it won't have swappable artifacts
        if (
            given_sagemaker_args.get("AlgorithmSpecification", {}).get("TrainingImage")
            is not None
        ):
            sagemaker_args = build_sagemaker_args(
                launch_project,
                self._api,
                role_arn,
                given_sagemaker_args.get("AlgorithmSpecification", {}).get(
                    "TrainingImage"
                ),
                default_output_path,
            )
            _logger.info(
                f"Launching sagemaker job on user supplied image with args: {sagemaker_args}"
            )
            run = launch_sagemaker_job(launch_project, sagemaker_args, sagemaker_client)
            if self.backend_config[PROJECT_SYNCHRONOUS]:
                run.wait()
            return run

        if launch_project.docker_image:
            image = launch_project.docker_image
        else:
            assert entry_point is not None
            assert builder is not None
            # build our own image
            _logger.info("Building docker image...")
            image = builder.build_image(
                launch_project,
                entry_point,
            )
            _logger.info(f"Docker image built with uri {image}")

        _logger.info("Connecting to sagemaker client")
        command_args = get_entry_point_command(
            entry_point, launch_project.override_args
        )
        if command_args:
            command_str = " ".join(command_args)
            wandb.termlog(
                f"{LOG_PREFIX}Launching run on sagemaker with entrypoint: {command_str}"
            )
        else:
            wandb.termlog(
                f"{LOG_PREFIX}Launching run on sagemaker with user-provided entrypoint in image"
            )
        sagemaker_args = build_sagemaker_args(
            launch_project, self._api, role_arn, image, default_output_path
        )
        _logger.info(f"Launching sagemaker job with args: {sagemaker_args}")
        run = launch_sagemaker_job(launch_project, sagemaker_args, sagemaker_client)
        if self.backend_config[PROJECT_SYNCHRONOUS]:
            run.wait()
        return run


def merge_aws_tag_with_algorithm_specification(
    algorithm_specification: Optional[Dict[str, Any]], aws_tag: Optional[str]
) -> Dict[str, Any]:
    """Create an AWS AlgorithmSpecification.

    AWS Sagemaker algorithms require a training image and an input mode. If the user
    does not specify the specification themselves, define the spec minimally using these
    two fields. Otherwise, if they specify the AlgorithmSpecification set the training
    image if it is not set.
    """
    if algorithm_specification is None:
        return {
            "TrainingImage": aws_tag,
            "TrainingInputMode": "File",
        }
    elif algorithm_specification.get("TrainingImage") is None:
        algorithm_specification["TrainingImage"] = aws_tag
    if algorithm_specification["TrainingImage"] is None:
        raise LaunchError("Failed determine tag for training image")
    return algorithm_specification


def build_sagemaker_args(
    launch_project: LaunchProject,
    api: Api,
    role_arn: str,
    aws_tag: Optional[str] = None,
    default_output_path: Optional[str] = None,
) -> Dict[str, Any]:
    sagemaker_args: Dict[str, Any] = {}
    given_sagemaker_args: Optional[Dict[str, Any]] = launch_project.resource_args.get(
        "sagemaker"
    )

    if given_sagemaker_args is None:
        raise LaunchError(
            "No sagemaker args specified. Specify sagemaker args in resource_args"
        )
    if (
        given_sagemaker_args.get("OutputDataConfig") is None
        and default_output_path is not None
    ):
        sagemaker_args["OutputDataConfig"] = {"S3OutputPath": default_output_path}
    else:
        sagemaker_args["OutputDataConfig"] = given_sagemaker_args.get(
            "OutputDataConfig"
        )

    if sagemaker_args.get("OutputDataConfig") is None:
        raise LaunchError(
            "Sagemaker launcher requires an OutputDataConfig Sagemaker resource argument"
        )
    training_job_name = cast(
        str, (given_sagemaker_args.get("TrainingJobName") or launch_project.run_id)
    )
    sagemaker_args["TrainingJobName"] = training_job_name

    sagemaker_args[
        "AlgorithmSpecification"
    ] = merge_aws_tag_with_algorithm_specification(
        given_sagemaker_args.get(
            "AlgorithmSpecification",
            given_sagemaker_args.get("algorithm_specification"),
        ),
        aws_tag,
    )

    sagemaker_args["RoleArn"] = role_arn

    camel_case_args = {
        to_camel_case(key): item for key, item in given_sagemaker_args.items()
    }
    sagemaker_args = {
        **camel_case_args,
        **sagemaker_args,
    }

    if sagemaker_args.get("ResourceConfig") is None:
        raise LaunchError(
            "Sagemaker launcher requires a ResourceConfig Sagemaker resource argument"
        )

    if sagemaker_args.get("StoppingCondition") is None:
        raise LaunchError(
            "Sagemaker launcher requires a StoppingCondition Sagemaker resource argument"
        )

    given_env = given_sagemaker_args.get(
        "Environment", sagemaker_args.get("environment", {})
    )
    calced_env = get_env_vars_dict(launch_project, api)
    total_env = {**calced_env, **given_env}
    sagemaker_args["Environment"] = total_env

    # remove args that were passed in for launch but not passed to sagemaker
    sagemaker_args.pop("EcrRepoName", None)
    sagemaker_args.pop("region", None)
    sagemaker_args.pop("profile", None)

    # clear the args that are None so they are not passed
    filtered_args = {k: v for k, v in sagemaker_args.items() if v is not None}

    return filtered_args


def launch_sagemaker_job(
    launch_project: LaunchProject,
    sagemaker_args: Dict[str, Any],
    sagemaker_client: "boto3.Client",
) -> SagemakerSubmittedRun:
    training_job_name = sagemaker_args.get("TrainingJobName") or launch_project.run_id
    resp = sagemaker_client.create_training_job(**sagemaker_args)

    if resp.get("TrainingJobArn") is None:
        raise LaunchError("Unable to create training job")

    run = SagemakerSubmittedRun(training_job_name, sagemaker_client)
    wandb.termlog(
        f"{LOG_PREFIX}Run job submitted with arn: {resp.get('TrainingJobArn')}"
    )
    url = "https://{region}.console.aws.amazon.com/sagemaker/home?region={region}#/jobs/{job_name}".format(
        region=sagemaker_client.meta.region_name, job_name=training_job_name
    )
    wandb.termlog(f"{LOG_PREFIX}See training job status at: {url}")
    return run


def get_role_arn(
    sagemaker_args: Dict[str, Any], backend_config: Dict[str, Any], account_id: str
) -> str:
    """Get the role arn from the sagemaker args or the backend config."""
    role_arn = sagemaker_args.get("RoleArn") or sagemaker_args.get("role_arn")
    if role_arn is None:
        role_arn = backend_config.get("runner", {}).get("role_arn")
    if role_arn is None or not isinstance(role_arn, str):
        raise LaunchError(
            "AWS sagemaker require a string RoleArn set this by adding a `RoleArn` key to the sagemaker"
            "field of resource_args"
        )
    if role_arn.startswith("arn:aws:iam::"):
        return role_arn

    return f"arn:aws:iam::{account_id}:role/{role_arn}"



import codecs
import json
import os
import sys
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Optional,
    Sequence,
    Set,
    TextIO,
    Tuple,
    Type,
    Union,
)

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict


import wandb
from wandb import util
from wandb.sdk.lib import runid

from . import _dtypes
from ._private import MEDIA_TMP
from .base_types.media import BatchableMedia

if TYPE_CHECKING:  # pragma: no cover
    import numpy as np  # type: ignore

    from ..wandb_artifacts import Artifact as LocalArtifact
    from ..wandb_run import Run as LocalRun

    numeric = Union[int, float, np.integer, np.float_]
    FileFormat3D = Literal[
        "obj",
        "gltf",
        "glb",
        "babylon",
        "stl",
        "pts.json",
    ]
    Point3D = Tuple[numeric, numeric, numeric]
    Point3DWithCategory = Tuple[numeric, numeric, numeric, numeric]
    Point3DWithColors = Tuple[numeric, numeric, numeric, numeric, numeric, numeric]
    Point = Union[Point3D, Point3DWithCategory, Point3DWithColors]
    PointCloudType = Literal["lidar/beta"]
    RGBColor = Tuple[int, int, int]

    class Box3D(TypedDict):
        corners: Tuple[
            Point3D,
            Point3D,
            Point3D,
            Point3D,
            Point3D,
            Point3D,
            Point3D,
            Point3D,
        ]
        label: Optional[str]
        color: RGBColor
        score: Optional[numeric]

    class Vector3D(TypedDict):
        start: Sequence[Point3D]
        end: Sequence[Point3D]

    class Camera(TypedDict):
        viewpoint: Sequence[Point3D]
        target: Sequence[Point3D]


class Object3D(BatchableMedia):
    """Wandb class for 3D point clouds.

    Arguments:
        data_or_path: (numpy array, string, io)
            Object3D can be initialized from a file or a numpy array.

            You can pass a path to a file or an io object and a file_type
            which must be one of SUPPORTED_TYPES

    The shape of the numpy array must be one of either:
    ```
    [[x y z],       ...] nx3
    [[x y z c],     ...] nx4 where c is a category with supported range [1, 14]
    [[x y z r g b], ...] nx6 where is rgb is color
    ```
    """

    SUPPORTED_TYPES: ClassVar[Set[str]] = {
        "obj",
        "gltf",
        "glb",
        "babylon",
        "stl",
        "pts.json",
    }
    SUPPORTED_POINT_CLOUD_TYPES: ClassVar[Set[str]] = {"lidar/beta"}
    _log_type: ClassVar[str] = "object3D-file"

    def __init__(
        self,
        data_or_path: Union["np.ndarray", str, "TextIO", dict],
        **kwargs: Optional[Union[str, "FileFormat3D"]],
    ) -> None:
        super().__init__()

        if hasattr(data_or_path, "name"):
            # if the file has a path, we just detect the type and copy it from there
            data_or_path = data_or_path.name

        if hasattr(data_or_path, "read"):
            if hasattr(data_or_path, "seek"):
                data_or_path.seek(0)
            object_3d = data_or_path.read()

            extension = kwargs.pop("file_type", None)
            if extension is None:
                raise ValueError(
                    "Must pass file type keyword argument when using io objects."
                )
            if extension not in Object3D.SUPPORTED_TYPES:
                raise ValueError(
                    "Object 3D only supports numpy arrays or files of the type: "
                    + ", ".join(Object3D.SUPPORTED_TYPES)
                )

            extension = "." + extension

            tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + extension)
            with open(tmp_path, "w") as f:
                f.write(object_3d)

            self._set_file(tmp_path, is_tmp=True, extension=extension)
        elif isinstance(data_or_path, str):
            path = data_or_path
            extension = None
            for supported_type in Object3D.SUPPORTED_TYPES:
                if path.endswith(supported_type):
                    extension = "." + supported_type
                    break

            if not extension:
                raise ValueError(
                    "File '"
                    + path
                    + "' is not compatible with Object3D: supported types are: "
                    + ", ".join(Object3D.SUPPORTED_TYPES)
                )

            self._set_file(data_or_path, is_tmp=False, extension=extension)
        # Supported different types and scene for 3D scenes
        elif isinstance(data_or_path, dict) and "type" in data_or_path:
            if data_or_path["type"] == "lidar/beta":
                data = {
                    "type": data_or_path["type"],
                    "vectors": data_or_path["vectors"].tolist()
                    if "vectors" in data_or_path
                    else [],
                    "points": data_or_path["points"].tolist()
                    if "points" in data_or_path
                    else [],
                    "boxes": data_or_path["boxes"].tolist()
                    if "boxes" in data_or_path
                    else [],
                }
            else:
                raise ValueError(
                    "Type not supported, only 'lidar/beta' is currently supported"
                )

            tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ".pts.json")
            with codecs.open(tmp_path, "w", encoding="utf-8") as fp:
                json.dump(
                    data,
                    fp,
                    separators=(",", ":"),
                    sort_keys=True,
                    indent=4,
                )
            self._set_file(tmp_path, is_tmp=True, extension=".pts.json")
        elif util.is_numpy_array(data_or_path):
            np_data = data_or_path

            # The following assertion is required for numpy to trust that
            # np_data is numpy array. The reason it is behind a False
            # guard is to ensure that this line does not run at runtime,
            # which would cause a runtime error if the user's machine did
            # not have numpy installed.

            if TYPE_CHECKING:
                assert isinstance(np_data, np.ndarray)

            if len(np_data.shape) != 2 or np_data.shape[1] not in {3, 4, 6}:
                raise ValueError(
                    """
                    The shape of the numpy array must be one of either
                                    [[x y z],       ...] nx3
                                     [x y z c],     ...] nx4 where c is a category with supported range [1, 14]
                                     [x y z r g b], ...] nx6 where rgb is color
                    """
                )

            list_data = np_data.tolist()
            tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ".pts.json")
            with codecs.open(tmp_path, "w", encoding="utf-8") as fp:
                json.dump(
                    list_data,
                    fp,
                    separators=(",", ":"),
                    sort_keys=True,
                    indent=4,
                )
            self._set_file(tmp_path, is_tmp=True, extension=".pts.json")
        else:
            raise ValueError("data must be a numpy array, dict or a file object")

    @classmethod
    def from_file(
        cls,
        data_or_path: Union["TextIO", str],
        file_type: Optional["FileFormat3D"] = None,
    ) -> "Object3D":
        """Initializes Object3D from a file or stream.

        Arguments:
            data_or_path (Union["TextIO", str]): A path to a file or a `TextIO` stream.
            file_type (str): Specifies the data format passed to `data_or_path`. Required when `data_or_path` is a
                `TextIO` stream. This parameter is ignored if a file path is provided. The type is taken from the file extension.
        """
        # if file_type is not None and file_type not in cls.SUPPORTED_TYPES:
        #     raise ValueError(
        #         f"Unsupported file type: {file_type}. Supported types are: {cls.SUPPORTED_TYPES}"
        #     )
        return cls(data_or_path, file_type=file_type)

    @classmethod
    def from_numpy(cls, data: "np.ndarray") -> "Object3D":
        """Initializes Object3D from a numpy array.

        Arguments:
            data (numpy array): Each entry in the array will
                represent one point in the point cloud.


        The shape of the numpy array must be one of either:
        ```
        [[x y z],       ...]  # nx3.
        [[x y z c],     ...]  # nx4 where c is a category with supported range [1, 14].
        [[x y z r g b], ...]  # nx6 where is rgb is color.
        ```
        """
        if not util.is_numpy_array(data):
            raise ValueError("`data` must be a numpy array")

        if len(data.shape) != 2 or data.shape[1] not in {3, 4, 6}:
            raise ValueError(
                """
                The shape of the numpy array must be one of either:
                                [[x y z],       ...] nx3
                                 [x y z c],     ...] nx4 where c is a category with supported range [1, 14]
                                 [x y z r g b], ...] nx6 where rgb is color
                """
            )

        return cls(data)

    @classmethod
    def from_point_cloud(
        cls,
        points: Sequence["Point"],
        boxes: Sequence["Box3D"],
        vectors: Optional[Sequence["Vector3D"]] = None,
        point_cloud_type: "PointCloudType" = "lidar/beta",
        # camera: Optional[Camera] = None,
    ) -> "Object3D":
        """Initializes Object3D from a python object.

        Arguments:
            points (Sequence["Point"]): The points in the point cloud.
            boxes (Sequence["Box3D"]): 3D bounding boxes for labeling the point cloud. Boxes
            are displayed in point cloud visualizations.
            vectors (Optional[Sequence["Vector3D"]]): Each vector is displayed in the point cloud
                visualization. Can be used to indicate directionality of bounding boxes. Defaults to None.
            point_cloud_type ("lidar/beta"): At this time, only the "lidar/beta" type is supported. Defaults to "lidar/beta".
        """
        if point_cloud_type not in cls.SUPPORTED_POINT_CLOUD_TYPES:
            raise ValueError("Point cloud type not supported")

        numpy = wandb.util.get_module(
            "numpy",
            required="wandb.Object3D.from_point_cloud requires numpy. Install with `pip install numpy`",
        )

        data = {
            "type": point_cloud_type,
            "points": numpy.array(points),
            "boxes": numpy.array(boxes),
            "vectors": numpy.array(vectors) if vectors is not None else numpy.array([]),
        }

        return cls(data)

    @classmethod
    def get_media_subdir(cls: Type["Object3D"]) -> str:
        return os.path.join("media", "object3D")

    def to_json(self, run_or_artifact: Union["LocalRun", "LocalArtifact"]) -> dict:
        json_dict = super().to_json(run_or_artifact)
        json_dict["_type"] = Object3D._log_type

        if isinstance(run_or_artifact, wandb.wandb_sdk.wandb_artifacts.Artifact):
            if self._path is None or not self._path.endswith(".pts.json"):
                raise ValueError(
                    "Non-point cloud 3D objects are not yet supported with Artifacts"
                )

        return json_dict

    @classmethod
    def seq_to_json(
        cls: Type["Object3D"],
        seq: Sequence["BatchableMedia"],
        run: "LocalRun",
        key: str,
        step: Union[int, str],
    ) -> dict:
        seq = list(seq)

        jsons = [obj.to_json(run) for obj in seq]

        for obj in jsons:
            expected = util.to_forward_slash_path(cls.get_media_subdir())
            if not obj["path"].startswith(expected):
                raise ValueError(
                    "Files in an array of Object3D's must be in the {} directory, not {}".format(
                        expected, obj["path"]
                    )
                )

        return {
            "_type": "object3D",
            "filenames": [
                os.path.relpath(j["path"], cls.get_media_subdir()) for j in jsons
            ],
            "count": len(jsons),
            "objects": jsons,
        }


class _Object3DFileType(_dtypes.Type):
    name = "object3D-file"
    types = [Object3D]


_dtypes.TypeRegistry.add(_Object3DFileType)



from pytest import raises

from promise import Promise
from promise.promise_list import PromiseList


def all(promises):
    return PromiseList(promises, Promise).promise


def test_empty_promises():
    all_promises = all([])
    assert all_promises.get() == []


def test_bad_promises():
    all_promises = all(None)

    with raises(Exception) as exc_info:
        all_promises.get()

    assert str(exc_info.value) == "PromiseList requires an iterable. Received None."


def test_promise_basic():
    all_promises = all([1, 2])
    assert all_promises.get() == [1, 2]


def test_promise_mixed():
    all_promises = all([1, 2, Promise.resolve(3)])
    assert all_promises.get() == [1, 2, 3]


def test_promise_rejected():
    e = Exception("Error")
    all_promises = all([1, 2, Promise.reject(e)])

    with raises(Exception) as exc_info:
        all_promises.get()

    assert str(exc_info.value) == "Error"


def test_promise_reject_skip_all_other_values():
    e1 = Exception("Error1")
    e2 = Exception("Error2")
    p = Promise()
    all_promises = all([1, Promise.reject(e1), Promise.reject(e2)])

    with raises(Exception) as exc_info:
        all_promises.get()

    assert str(exc_info.value) == "Error1"


def test_promise_lazy_promise():
    p = Promise()
    all_promises = all([1, 2, p])
    assert not all_promises.is_fulfilled
    p.do_resolve(3)
    assert all_promises.get() == [1, 2, 3]


def test_promise_contained_promise():
    p = Promise()
    all_promises = all([1, 2, Promise.resolve(None).then(lambda v: p)])
    assert not all_promises.is_fulfilled
    p.do_resolve(3)
    assert all_promises.get() == [1, 2, 3]



"""Implementation of the GCP environment for wandb launch."""
import logging
import os
import re

from wandb.sdk.launch.utils import LaunchError
from wandb.util import get_module

from .abstract import AbstractEnvironment

google = get_module(
    "google",
    required="Google Cloud Platform support requires the google package. Please"
    " install it with `pip install wandb[launch]`.",
)
google.cloud.compute_v1 = get_module(
    "google.cloud.compute_v1",
    required="Google Cloud Platform support requires the google-cloud-compute package. "
    "Please install it with `pip install wandb[launch]`.",
)
google.auth.credentials = get_module(
    "google.auth.credentials",
    required="Google Cloud Platform support requires google-auth. "
    "Please install it with `pip install wandb[launch]`.",
)
google.auth.transport.requests = get_module(
    "google.auth.transport.requests",
    required="Google Cloud Platform support requires google-auth. "
    "Please install it with `pip install wandb[launch]`.",
)
google.api_core.exceptions = get_module(
    "google.api_core.exceptions",
    required="Google Cloud Platform support requires google-api-core. "
    "Please install it with `pip install wandb[launch]`.",
)
google.cloud.storage = get_module(
    "google.cloud.storage",
    required="Google Cloud Platform support requires google-cloud-storage. "
    "Please install it with `pip install wandb[launch].",
)


_logger = logging.getLogger(__name__)

GCS_URI_RE = re.compile(r"gs://([^/]+)/(.+)")


class GcpEnvironment(AbstractEnvironment):
    """GCP Environment.

    Attributes:
        region: The GCP region.
    """

    region: str

    def __init__(self, region: str, verify: bool = True) -> None:
        """Initialize the GCP environment.

        Arguments:
            region: The GCP region.
            verify: Whether to verify the credentials, region, and project.

        Raises:
            LaunchError: If verify is True and the environment is not properly
                configured.
        """
        super().__init__()
        _logger.info(f"Initializing GcpEnvironment in region {region}")
        self.region: str = region
        self._project = ""
        if verify:
            self.verify()

    @classmethod
    def from_config(cls, config: dict) -> "GcpEnvironment":
        """Create a GcpEnvironment from a config dictionary.

        Arguments:
            config: The config dictionary.

        Returns:
            GcpEnvironment: The GcpEnvironment.
        """
        if config.get("type") != "gcp":
            raise LaunchError(
                f"Could not create GcpEnvironment from config. Expected type 'gcp' "
                f"but got '{config.get('type')}'."
            )
        region = config.get("region", None)
        if not region:
            raise LaunchError(
                "Could not create GcpEnvironment from config. Missing 'region' "
                "field."
            )
        return cls(region=region)

    @property
    def project(self) -> str:
        """Get the name of the gcp project.

        The project name is determined by the credentials, so this method
        verifies the credentials if they have not already been verified.

        Returns:
            str: The name of the gcp project.

        Raises:
            LaunchError: If the launch environment cannot be verified.
        """
        if not self._project:
            raise LaunchError(
                "This GcpEnvironment has not been verified. Please call verify() "
                "before accessing the project."
            )
        return self._project

    def get_credentials(self) -> google.auth.credentials.Credentials:  # type: ignore
        """Get the GCP credentials.

        Uses google.auth.default() to get the credentials. If the credentials
        are invalid, this method will refresh them. If the credentials are
        still invalid after refreshing, this method will raise an error.

        Returns:
            google.auth.credentials.Credentials: The GCP credentials.

        Raises:
            LaunchError: If the GCP credentials are invalid.
        """
        _logger.debug("Getting GCP credentials")
        # TODO: Figure out a minimal set of scopes.
        scopes = [
            "https://www.googleapis.com/auth/cloud-platform",
        ]
        try:
            creds, project = google.auth.default(scopes=scopes)
            if not self._project:
                self._project = project
            _logger.debug("Refreshing GCP credentials")
            creds.refresh(google.auth.transport.requests.Request())
        except google.auth.exceptions.DefaultCredentialsError as e:
            raise LaunchError(
                "No Google Cloud Platform credentials found. Please run "
                "`gcloud auth application-default login` or set the environment "
                "variable GOOGLE_APPLICATION_CREDENTIALS to the path of a valid "
                "service account key file."
            ) from e
        except google.auth.exceptions.RefreshError as e:
            raise LaunchError(
                "Could not refresh Google Cloud Platform credentials. Please run "
                "`gcloud auth application-default login` or set the environment "
                "variable GOOGLE_APPLICATION_CREDENTIALS to the path of a valid "
                "service account key file."
            ) from e
        if not creds.valid:
            raise LaunchError(
                "Invalid Google Cloud Platform credentials. Please run "
                "`gcloud auth application-default login` or set the environment "
                "variable GOOGLE_APPLICATION_CREDENTIALS to the path of a valid "
                "service account key file."
            )
        return creds

    def verify(self) -> None:
        """Verify the credentials, region, and project.

        Credentials and region are verified by calling get_credentials(). The
        region and is verified by calling the compute API.

        Raises:
            LaunchError: If the credentials, region, or project are invalid.

        Returns:
            None
        """
        _logger.debug("Verifying GCP environment")
        creds = self.get_credentials()
        try:
            # Check if the region is available using the compute API.
            compute_client = google.cloud.compute_v1.RegionsClient(credentials=creds)
            compute_client.get(project=self.project, region=self.region)
        except google.api_core.exceptions.NotFound as e:
            raise LaunchError(
                f"Region {self.region} is not available in project {self.project}."
            ) from e

    def verify_storage_uri(self, uri: str) -> None:
        """Verify that a storage URI is valid.

        Arguments:
            uri: The storage URI.

        Raises:
            LaunchError: If the storage URI is invalid.
        """
        match = GCS_URI_RE.match(uri)
        if not match:
            raise LaunchError(f"Invalid GCS URI: {uri}")
        bucket = match.group(1)
        try:
            storage_client = google.cloud.storage.Client(
                credentials=self.get_credentials()
            )
            bucket = storage_client.post_bucket(bucket)
        except google.api_core.exceptions.NotFound as e:
            raise LaunchError(f"Bucket {bucket} does not exist.") from e

    def upload_file(self, source: str, destination: str) -> None:
        """Upload a file to GCS.

        Arguments:
            source: The path to the local file.
            destination: The path to the GCS file.

        Raises:
            LaunchError: If the file cannot be uploaded.
        """
        _logger.debug(f"Uploading file {source} to {destination}")
        if not os.path.isfile(source):
            raise LaunchError(f"File {source} does not exist.")
        match = GCS_URI_RE.match(destination)
        if not match:
            raise LaunchError(f"Invalid GCS URI: {destination}")
        bucket = match.group(1)
        key = match.group(2).lstrip("/")
        try:
            storage_client = google.cloud.storage.Client(
                credentials=self.get_credentials()
            )
            bucket = storage_client.bucket(bucket)
            blob = bucket.blob(key)
            blob.upload_from_filename(source)
        except google.api_core.exceptions.GoogleAPICallError as e:
            raise LaunchError(f"Could not upload file to GCS: {e}") from e

    def upload_dir(self, source: str, destination: str) -> None:
        """Upload a directory to GCS.

        Arguments:
            source: The path to the local directory.
            destination: The path to the GCS directory.

        Raises:
            LaunchError: If the directory cannot be uploaded.
        """
        _logger.debug(f"Uploading directory {source} to {destination}")
        if not os.path.isdir(source):
            raise LaunchError(f"Directory {source} does not exist.")
        match = GCS_URI_RE.match(destination)
        if not match:
            raise LaunchError(f"Invalid GCS URI: {destination}")
        bucket = match.group(1)
        key = match.group(2).lstrip("/")
        try:
            storage_client = google.cloud.storage.Client(
                credentials=self.get_credentials()
            )
            bucket = storage_client.bucket(bucket)
            for root, _, files in os.walk(source):
                for file in files:
                    local_path = os.path.join(root, file)
                    gcs_path = os.path.join(
                        key, os.path.relpath(local_path, source)
                    ).replace("\\", "/")
                    blob = bucket.blob(gcs_path)
                    blob.upload_from_filename(local_path)
        except google.api_core.exceptions.GoogleAPICallError as e:
            raise LaunchError(f"Could not upload directory to GCS: {e}") from e
            raise LaunchError(f"Could not upload directory to GCS: {e}") from e
            raise LaunchError(f"Could not upload directory to GCS: {e}") from e



from typing import Any, Dict, Tuple

from wandb.data_types import Table
from wandb.errors import Error


class Visualize:
    def __init__(self, id: str, data: Table) -> None:
        self._id = id
        self._data = data

    def get_config_value(self, key: str) -> Dict[str, Any]:
        return {
            "id": self._id,
            "historyFieldSettings": {"x-axis": "_step", "key": key},
        }

    @staticmethod
    def get_config_key(key: str) -> Tuple[str, str, str]:
        return "_wandb", "viz", key

    @property
    def value(self) -> Table:
        return self._data


class CustomChart:
    def __init__(
        self,
        id: str,
        data: Table,
        fields: Dict[str, Any],
        string_fields: Dict[str, Any],
    ) -> None:
        self._id = id
        self._data = data
        self._fields = fields
        self._string_fields = string_fields

    def get_config_value(
        self,
        panel_type: str,
        query: Dict[str, Any],
    ) -> Dict[str, Any]:
        return {
            "panel_type": panel_type,
            "panel_config": {
                "panelDefId": self._id,
                "fieldSettings": self._fields,
                "stringSettings": self._string_fields,
                "transform": {"name": "tableWithLeafColNames"},
                "userQuery": query,
            },
        }

    @staticmethod
    def get_config_key(key: str) -> Tuple[str, str, str]:
        return "_wandb", "visualize", key

    @staticmethod
    def user_query(table_key: str) -> Dict[str, Any]:
        return {
            "queryFields": [
                {
                    "name": "runSets",
                    "args": [{"name": "runSets", "value": "${runSets}"}],
                    "fields": [
                        {"name": "id", "fields": []},
                        {"name": "name", "fields": []},
                        {"name": "_defaultColorIndex", "fields": []},
                        {
                            "name": "summaryTable",
                            "args": [{"name": "tableKey", "value": table_key}],
                            "fields": [],
                        },
                    ],
                }
            ],
        }

    @property
    def table(self) -> Table:
        return self._data

    @property
    def fields(self) -> Dict[str, Any]:
        return self._fields

    @property
    def string_fields(self) -> Dict[str, Any]:
        return self._string_fields


def custom_chart(
    vega_spec_name: str,
    data_table: Table,
    fields: Dict[str, Any],
    string_fields: Dict[str, Any] = {},
) -> CustomChart:
    if not isinstance(data_table, Table):
        raise Error(
            f"Expected `data_table` to be `wandb.Table` type, instead got {type(data_table).__name__}"
        )
    return CustomChart(
        id=vega_spec_name,
        data=data_table,
        fields=fields,
        string_fields=string_fields,
    )


def visualize(id: str, value: Table) -> Visualize:
    if not isinstance(value, Table):
        raise Error(
            f"Expected `value` to be `wandb.Table` type, instead got {type(value).__name__}"
        )
    return Visualize(id=id, data=value)



"""Reliably launch and connect to backend server process (wandb service).

Backend server process can be connected to using tcp sockets or grpc transport.
"""

import os
import platform
import shutil
import subprocess
import sys
import tempfile
import time
from typing import TYPE_CHECKING, Any, Dict, Optional

from wandb import _sentry
from wandb.errors import Error

from . import _startup_debug, port_file
from .service_base import ServiceInterface
from .service_sock import ServiceSockInterface

if TYPE_CHECKING:
    from wandb.sdk.wandb_settings import Settings


class ServiceStartProcessError(Error):
    """Raised when a known error occurs when launching wandb service."""

    pass


class ServiceStartTimeoutError(Error):
    """Raised when service start times out."""

    pass


class ServiceStartPortError(Error):
    """Raised when service start fails to find a port."""

    pass


class _Service:
    _settings: "Settings"
    _grpc_port: Optional[int]
    _sock_port: Optional[int]
    _service_interface: ServiceInterface
    _internal_proc: Optional[subprocess.Popen]
    _use_grpc: bool
    _startup_debug_enabled: bool

    def __init__(
        self,
        settings: "Settings",
    ) -> None:
        self._settings = settings
        self._stub = None
        self._grpc_port = None
        self._sock_port = None
        self._internal_proc = None
        self._startup_debug_enabled = _startup_debug.is_enabled()

        _sentry.configure_scope(process_context="service")

        # Temporary setting to allow use of grpc so that we can keep
        # that code from rotting during the transition
        self._use_grpc = self._settings._service_transport == "grpc"

        # current code only supports grpc or socket server implementation, in the
        # future we might be able to support both
        if self._use_grpc:
            from .service_grpc import ServiceGrpcInterface

            self._service_interface = ServiceGrpcInterface()
        else:
            self._service_interface = ServiceSockInterface()

    def _startup_debug_print(self, message: str) -> None:
        if not self._startup_debug_enabled:
            return
        _startup_debug.print_message(message)

    def _wait_for_ports(
        self, fname: str, proc: Optional[subprocess.Popen] = None
    ) -> None:
        """Wait for the service to write the port file and then read it.

        Args:
            fname: The path to the port file.
            proc: The process to wait for.

        Raises:
            ServiceStartTimeoutError: If the service takes too long to start.
            ServiceStartPortError: If the service writes an invalid port file or unable to read it.
            ServiceStartProcessError: If the service process exits unexpectedly.

        """
        time_max = time.monotonic() + self._settings._service_wait
        while time.monotonic() < time_max:
            if proc and proc.poll():
                # process finished
                # define these variables for sentry context grab:
                # command = proc.args
                # sys_executable = sys.executable
                # which_python = shutil.which("python3")
                # proc_out = proc.stdout.read()
                # proc_err = proc.stderr.read()
                context = dict(
                    command=proc.args,
                    sys_executable=sys.executable,
                    which_python=shutil.which("python3"),
                    proc_out=proc.stdout.read() if proc.stdout else "",
                    proc_err=proc.stderr.read() if proc.stderr else "",
                )
                raise ServiceStartProcessError(
                    f"The wandb service process exited with {proc.returncode}. "
                    "Ensure that `sys.executable` is a valid python interpreter. "
                    "You can override it with the `_executable` setting "
                    "or with the `WANDB__EXECUTABLE` environment variable.",
                    context=context,
                )
            if not os.path.isfile(fname):
                time.sleep(0.2)
                continue
            try:
                pf = port_file.PortFile()
                pf.read(fname)
                if not pf.is_valid:
                    time.sleep(0.2)
                    continue
                self._grpc_port = pf.grpc_port
                self._sock_port = pf.sock_port
            except Exception as e:
                # todo: point at the docs. this could be due to a number of reasons,
                #  for example, being unable to write to the port file etc.
                raise ServiceStartPortError(
                    f"Failed to allocate port for wandb service: {e}."
                )
            return
        raise ServiceStartTimeoutError(
            "Timed out waiting for wandb service to start after "
            f"{self._settings._service_wait} seconds. "
            "Try increasing the timeout with the `_service_wait` setting."
        )

    def _launch_server(self) -> None:
        """Launch server and set ports."""
        # References for starting processes
        # - https://github.com/wandb/wandb/blob/archive/old-cli/wandb/__init__.py
        # - https://stackoverflow.com/questions/1196074/how-to-start-a-background-process-in-python
        self._startup_debug_print("launch")

        kwargs: Dict[str, Any] = dict(close_fds=True)
        # flags to handle keyboard interrupt signal that is causing a hang
        if platform.system() == "Windows":
            kwargs.update(creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)  # type: ignore [attr-defined]
        else:
            kwargs.update(start_new_session=True)

        pid = str(os.getpid())

        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, f"port-{pid}.txt")

            executable = self._settings._executable
            exec_cmd_list = [executable, "-m"]
            # Add coverage collection if needed
            if os.environ.get("YEA_RUN_COVERAGE") and os.environ.get("COVERAGE_RCFILE"):
                exec_cmd_list += ["coverage", "run", "-m"]
            service_args = [
                "wandb",
                "service",
                "--port-filename",
                fname,
                "--pid",
                pid,
                "--debug",
            ]
            if self._use_grpc:
                service_args.append("--serve-grpc")
            else:
                service_args.append("--serve-sock")
            internal_proc = subprocess.Popen(
                exec_cmd_list + service_args,
                env=os.environ,
                **kwargs,
            )
            self._startup_debug_print("wait_ports")
            try:
                self._wait_for_ports(fname, proc=internal_proc)
            except Exception as e:
                _sentry.reraise(e)
            self._startup_debug_print("wait_ports_done")
            self._internal_proc = internal_proc
        self._startup_debug_print("launch_done")

    def start(self) -> None:
        self._launch_server()

    @property
    def grpc_port(self) -> Optional[int]:
        return self._grpc_port

    @property
    def sock_port(self) -> Optional[int]:
        return self._sock_port

    @property
    def service_interface(self) -> ServiceInterface:
        return self._service_interface

    def join(self) -> int:
        ret = 0
        if self._internal_proc:
            ret = self._internal_proc.wait()
        return ret



from warnings import simplefilter

import wandb

# ignore all future warnings
simplefilter(action="ignore", category=FutureWarning)


def decision_boundaries(
    decision_boundary_x,
    decision_boundary_y,
    decision_boundary_color,
    train_x,
    train_y,
    train_color,
    test_x,
    test_y,
    test_color,
):
    x_dict, y_dict, color_dict = [], [], []
    for i in range(min(len(decision_boundary_x), 100)):
        x_dict.append(decision_boundary_x[i])
        y_dict.append(decision_boundary_y[i])
        color_dict.append(decision_boundary_color)
    for i in range(300):
        x_dict.append(test_x[i])
        y_dict.append(test_y[i])
        color_dict.append(test_color[i])
    for i in range(min(len(train_x), 600)):
        x_dict.append(train_x[i])
        y_dict.append(train_y[i])
        color_dict.append(train_color[i])

    return wandb.visualize(
        "wandb/decision_boundaries/v1",
        wandb.Table(
            columns=["x", "y", "color"],
            data=[[x_dict[i], y_dict[i], color_dict[i]] for i in range(len(x_dict))],
        ),
    )



from . import ast

QUERY_DOCUMENT_KEYS = {
    ast.Name: (),

    ast.Document: ('definitions',),
    ast.OperationDefinition: ('name', 'variable_definitions', 'directives', 'selection_set'),
    ast.VariableDefinition: ('variable', 'type', 'default_value'),
    ast.Variable: ('name',),
    ast.SelectionSet: ('selections',),
    ast.Field: ('alias', 'name', 'arguments', 'directives', 'selection_set'),
    ast.Argument: ('name', 'value'),

    ast.FragmentSpread: ('name', 'directives'),
    ast.InlineFragment: ('type_condition', 'directives', 'selection_set'),
    ast.FragmentDefinition: ('name', 'type_condition', 'directives', 'selection_set'),

    ast.IntValue: (),
    ast.FloatValue: (),
    ast.StringValue: (),
    ast.BooleanValue: (),
    ast.EnumValue: (),
    ast.ListValue: ('values',),
    ast.ObjectValue: ('fields',),
    ast.ObjectField: ('name', 'value'),

    ast.Directive: ('name', 'arguments'),

    ast.NamedType: ('name',),
    ast.ListType: ('type',),
    ast.NonNullType: ('type',),

    ast.SchemaDefinition: ('directives', 'operation_types',),
    ast.OperationTypeDefinition: ('type',),

    ast.ScalarTypeDefinition: ('name', 'directives',),
    ast.ObjectTypeDefinition: ('name', 'interfaces', 'directives', 'fields'),
    ast.FieldDefinition: ('name', 'arguments', 'directives', 'type'),
    ast.InputValueDefinition: ('name', 'type', 'directives', 'default_value'),
    ast.InterfaceTypeDefinition: ('name', 'directives', 'fields'),
    ast.UnionTypeDefinition: ('name', 'directives', 'types'),
    ast.EnumTypeDefinition: ('name', 'directives', 'values'),
    ast.EnumValueDefinition: ('name', 'directives',),
    ast.InputObjectTypeDefinition: ('name', 'directives', 'fields'),

    ast.TypeExtensionDefinition: ('definition',),

    ast.DirectiveDefinition: ('name', 'arguments', 'locations'),
}

AST_KIND_TO_TYPE = {c.__name__: c for c in QUERY_DOCUMENT_KEYS.keys()}


class VisitorMeta(type):

    def __new__(cls, name, bases, attrs):
        enter_handlers = {}
        leave_handlers = {}

        for base in bases:
            if hasattr(base, '_enter_handlers'):
                enter_handlers.update(base._enter_handlers)

            if hasattr(base, '_leave_handlers'):
                leave_handlers.update(base._leave_handlers)

        for attr, val in attrs.items():
            if attr.startswith('enter_'):
                ast_kind = attr[6:]
                ast_type = AST_KIND_TO_TYPE.get(ast_kind)
                enter_handlers[ast_type] = val

            elif attr.startswith('leave_'):
                ast_kind = attr[6:]
                ast_type = AST_KIND_TO_TYPE.get(ast_kind)
                leave_handlers[ast_type] = val

        attrs['_enter_handlers'] = enter_handlers
        attrs['_leave_handlers'] = leave_handlers
        attrs['_get_enter_handler'] = enter_handlers.get
        attrs['_get_leave_handler'] = leave_handlers.get
        return super(VisitorMeta, cls).__new__(cls, name, bases, attrs)



"""Tools for integrating `wandb` with [`ultralytics YOLOv8`](https://docs.ultralytics.com/).

YOLOv8 is a computer vision framework for training and deploying YOLOv8 models.
"""
__all__ = ("add_callbacks",)

from .yolov8 import add_callbacks



from typing import Iterator, List, Type

from .interfaces import Asset


class AssetRegistry:
    def __init__(self) -> None:
        self._registry: List[Type[Asset]] = []

    def register(self, asset: Type[Asset]) -> Type[Asset]:
        self._registry.append(asset)
        return asset

    def __iter__(self) -> Iterator[Type[Asset]]:
        for asset in self._registry:
            if asset.is_available():
                yield asset


asset_registry = AssetRegistry()



import json
import platform
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor, as_completed
from contextlib import contextmanager
from typing import Any, Dict, Iterable, List, Optional, Tuple

from tqdm.auto import tqdm

import wandb
from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_telemetry_pb2 as telem_pb
from wandb.sdk.interface.interface import file_policy_to_enum
from wandb.sdk.interface.interface_queue import InterfaceQueue
from wandb.sdk.internal.sender import SendManager

Name = str
Path = str


def coalesce(*arg: Any) -> Any:
    """Return the first non-none value in the list of arguments.  Similar to ?? in C#."""
    return next((a for a in arg if a is not None), None)


@contextmanager
def send_manager(root_dir):
    sm = SendManager.setup(root_dir, resume=False)
    try:
        yield sm
    finally:
        # flush any remaining records
        while sm:
            data = next(sm)
            sm.send(data)
        sm.finish()


class ImporterRun:
    def __init__(self) -> None:
        self.interface = InterfaceQueue()
        self.run_dir = f"./wandb-importer/{self.run_id()}"

    def run_id(self) -> str:
        _id = wandb.util.generate_id()
        wandb.termwarn(f"`run_id` not specified.  Autogenerating id: {_id}")
        return _id

    def entity(self) -> str:
        _entity = "unspecified-entity"
        wandb.termwarn(f"`entity` not specified.  Defaulting to: {_entity}")
        return _entity

    def project(self) -> str:
        _project = "unspecified-project"
        wandb.termwarn(f"`project` not specified.  Defaulting to: {_project}")
        return _project

    def config(self) -> Dict[str, Any]:
        return {}

    def summary(self) -> Dict[str, float]:
        return {}

    def metrics(self) -> List[Dict[str, float]]:
        """Metrics for the run.

        We expect metrics in this shape:

        [
            {'metric1': 1, 'metric2': 1, '_step': 0},
            {'metric1': 2, 'metric2': 4, '_step': 1},
            {'metric1': 3, 'metric2': 9, '_step': 2},
            ...
        ]

        You can also submit metrics in this shape:
        [
            {'metric1': 1, '_step': 0},
            {'metric2': 1, '_step': 0},
            {'metric1': 2, '_step': 1},
            {'metric2': 4, '_step': 1},
            ...
        ]
        """
        return []

    def run_group(self) -> Optional[str]:
        ...

    def job_type(self) -> Optional[str]:
        ...

    def display_name(self) -> str:
        return self.run_id()

    def notes(self) -> Optional[str]:
        ...

    def tags(self) -> Optional[List[str]]:
        ...

    def artifacts(self) -> Optional[Iterable[Tuple[Name, Path]]]:
        ...

    def os_version(self) -> Optional[str]:
        ...

    def python_version(self) -> Optional[str]:
        ...

    def cuda_version(self) -> Optional[str]:
        ...

    def program(self) -> Optional[str]:
        ...

    def host(self) -> Optional[str]:
        ...

    def username(self) -> Optional[str]:
        ...

    def executable(self) -> Optional[str]:
        ...

    def gpus_used(self) -> Optional[str]:
        ...

    def cpus_used(self) -> Optional[int]:  # can we get the model?
        ...

    def memory_used(self) -> Optional[int]:
        ...

    def runtime(self) -> Optional[int]:
        ...

    def start_time(self) -> Optional[int]:
        ...

    def _make_run_record(self) -> pb.Record:
        run = pb.RunRecord()
        run.run_id = self.run_id()
        run.entity = self.entity()
        run.project = self.project()
        run.display_name = coalesce(self.display_name())
        run.notes = coalesce(self.notes(), "")
        run.tags.extend(coalesce(self.tags(), list()))
        # run.start_time.FromMilliseconds(self.start_time())
        # run.runtime = self.runtime()
        run_group = self.run_group()
        if run_group is not None:
            run.run_group = run_group
        self.interface._make_config(
            data=self.config(),
            obj=run.config,
        )  # is there a better way?
        return self.interface._make_record(run=run)

    def _make_summary_record(self) -> pb.Record:
        d: dict = {
            **self.summary(),
            "_runtime": self.runtime(),  # quirk of runtime -- it has to be here!
            # '_timestamp': self.start_time()/1000,
        }
        summary = self.interface._make_summary_from_dict(d)
        return self.interface._make_record(summary=summary)

    def _make_history_records(self) -> Iterable[pb.Record]:
        for _, metrics in enumerate(self.metrics()):
            history = pb.HistoryRecord()
            for k, v in metrics.items():
                item = history.item.add()
                item.key = k
                item.value_json = json.dumps(v)
            yield self.interface._make_record(history=history)

    def _make_files_record(self, files_dict) -> pb.Record:
        # when making the metadata file, it captures most things correctly
        # but notably it doesn't capture the start time!
        files_record = pb.FilesRecord()
        for path, policy in files_dict["files"]:
            f = files_record.files.add()
            f.path = path
            f.policy = file_policy_to_enum(policy)  # is this always "end"?
        return self.interface._make_record(files=files_record)

    def _make_metadata_files_record(self) -> pb.Record:
        self._make_metadata_file(self.run_dir)
        return self._make_files_record(
            {"files": [[f"{self.run_dir}/files/wandb-metadata.json", "end"]]}
        )

    def _make_artifact_record(self) -> pb.Record:
        art = wandb.Artifact(self.display_name(), "imported-artifacts")
        artifacts = self.artifacts()
        if artifacts is not None:
            for name, path in artifacts:
                art.add_file(path, name)
        proto = self.interface._make_artifact(art)
        proto.run_id = self.run_id()
        proto.project = self.project()
        proto.entity = self.entity()
        proto.user_created = False
        proto.use_after_commit = False
        proto.finalize = True
        for tag in ["latest", "imported"]:
            proto.aliases.append(tag)
        return self.interface._make_record(artifact=proto)

    def _make_telem_record(self) -> pb.Record:
        feature = telem_pb.Feature()
        feature.importer_mlflow = True

        telem = telem_pb.TelemetryRecord()
        telem.feature.CopyFrom(feature)
        telem.python_version = platform.python_version()  # importer's python version
        telem.cli_version = wandb.__version__
        return self.interface._make_record(telemetry=telem)

    def _make_metadata_file(self, run_dir: str) -> None:
        missing_text = "MLFlow did not capture this info."

        d = {}
        if self.os_version() is not None:
            d["os"] = self.os_version()
        else:
            d["os"] = missing_text

        if self.python_version() is not None:
            d["python"] = self.python_version()
        else:
            d["python"] = missing_text

        if self.program() is not None:
            d["program"] = self.program()
        else:
            d["program"] = missing_text

        if self.cuda_version() is not None:
            d["cuda"] = self.cuda_version()
        if self.host() is not None:
            d["host"] = self.host()
        if self.username() is not None:
            d["username"] = self.username()
        if self.executable() is not None:
            d["executable"] = self.executable()
        gpus_used = self.gpus_used()
        if gpus_used is not None:
            d["gpu_devices"] = json.dumps(gpus_used)
            d["gpu_count"] = json.dumps(len(gpus_used))
        cpus_used = self.cpus_used()
        if cpus_used is not None:
            d["cpu_count"] = json.dumps(self.cpus_used())
        mem_used = self.memory_used()
        if mem_used is not None:
            d["memory"] = json.dumps({"total": self.memory_used()})

        with open(f"{run_dir}/files/wandb-metadata.json", "w") as f:
            f.write(json.dumps(d))


class Importer(ABC):
    @abstractmethod
    def download_all_runs(self) -> Iterable[ImporterRun]:
        ...

    def import_all(self, overrides: Optional[Dict[str, Any]] = None) -> None:
        for run in tqdm(self.download_all_runs(), desc="Sending runs"):
            self.import_one(run, overrides)

    def import_all_parallel(
        self, overrides: Optional[Dict[str, Any]] = None, **pool_kwargs: Any
    ) -> None:
        runs = list(self.download_all_runs())
        with tqdm(total=len(runs)) as pbar:
            with ProcessPoolExecutor(**pool_kwargs) as exc:
                futures = {
                    exc.submit(self.import_one, run, overrides=overrides): run
                    for run in runs
                }
                for future in as_completed(futures):
                    run = futures[future]
                    pbar.update(1)
                    pbar.set_description(
                        f"Imported Run: {run.run_group()} {run.display_name()}"
                    )

    def import_one(
        self,
        run: ImporterRun,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        # does this need to be here for pmap?
        if overrides:
            for k, v in overrides.items():
                # `lambda: v` won't work!
                # https://stackoverflow.com/questions/10802002/why-deepcopy-doesnt-create-new-references-to-lambda-function
                setattr(run, k, lambda v=v: v)
        self._import_one(run)

    def _import_one(self, run: ImporterRun) -> None:
        with send_manager(run.run_dir) as sm:
            sm.send(run._make_run_record())
            sm.send(run._make_summary_record())
            sm.send(run._make_metadata_files_record())
            for history_record in run._make_history_records():
                sm.send(history_record)
            if run.artifacts() is not None:
                sm.send(run._make_artifact_record())
            sm.send(run._make_telem_record())



"""console."""

import os


def win32_redirect(stdout_slave_fd, stderr_slave_fd):
    # import win32api

    # save for later
    # fd_stdout = os.dup(1)
    # fd_stderr = os.dup(2)

    # std_out = win32api.GetStdHandle(win32api.STD_OUTPUT_HANDLE)
    # std_err = win32api.GetStdHandle(win32api.STD_ERROR_HANDLE)

    # os.dup2(stdout_slave_fd, 1)
    # os.dup2(stderr_slave_fd, 2)

    # TODO(jhr): do something about current stdout, stderr file handles
    pass


def win32_create_pipe():
    # import pywintypes
    # import win32pipe

    # sa=pywintypes.SECURITY_ATTRIBUTES()
    # sa.bInheritHandle=1

    # read_fd, write_fd = win32pipe.FdCreatePipe(sa, 0, os.O_TEXT)
    # read_fd, write_fd = win32pipe.FdCreatePipe(sa, 0, os.O_BINARY)
    read_fd, write_fd = os.pipe()
    # http://timgolden.me.uk/pywin32-docs/win32pipe__FdCreatePipe_meth.html
    # https://stackoverflow.com/questions/17942874/stdout-redirection-with-ctypes

    # f = open("testing.txt", "rb")
    # read_fd = f.fileno()

    return read_fd, write_fd



import io
import re
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import wandb
import wandb.util
from wandb.sdk.lib import telemetry
from wandb.viz import custom_chart

if TYPE_CHECKING:
    import numpy as np  # type: ignore

    from wandb.sdk.internal.tb_watcher import TBHistory

# We have at least the default namestep and a global step to track
# TODO: reset this structure on wandb.finish
STEPS: Dict[str, Dict[str, Any]] = {
    "": {"step": 0},
    "global": {"step": 0, "last_log": None},
}
# TODO(cling): Set these when tensorboard behavior is configured.
# We support rate limited logging by setting this to number of seconds,
# can be a floating point.
RATE_LIMIT_SECONDS: Optional[Union[float, int]] = None
IGNORE_KINDS = ["graphs"]
tensor_util = wandb.util.get_module("tensorboard.util.tensor_util")


# prefer tensorboard, fallback to protobuf in tensorflow when tboard isn't available
pb = wandb.util.get_module(
    "tensorboard.compat.proto.summary_pb2"
) or wandb.util.get_module("tensorflow.core.framework.summary_pb2")

Summary = pb.Summary if pb else None


def make_ndarray(tensor: Any) -> Optional["np.ndarray"]:
    if tensor_util:
        res = tensor_util.make_ndarray(tensor)
        # Tensorboard can log generic objects, and we don't want to save them
        if res.dtype == "object":
            return None
        else:
            return res
    else:
        wandb.termwarn(
            "Can't convert tensor summary, upgrade tensorboard with `pip"
            " install tensorboard --upgrade`"
        )
        return None


def namespaced_tag(tag: str, namespace: str = "") -> str:
    if not namespace:
        return tag
    elif tag in namespace:
        # This happens with tensorboardX
        return namespace
    else:
        return namespace + "/" + tag


def history_image_key(key: str, namespace: str = "") -> str:
    """Convert invalid filesystem characters to _ for use in History keys.

    Unfortunately this means currently certain image keys will collide silently. We
    implement this mapping up here in the TensorFlow stuff rather than in the History
    stuff so that we don't have to store a mapping anywhere from the original keys to
    the safe ones.
    """
    return namespaced_tag(re.sub(r"[/\\]", "_", key), namespace)


def tf_summary_to_dict(  # noqa: C901
    tf_summary_str_or_pb: Any, namespace: str = ""
) -> Optional[Dict[str, Any]]:
    """Convert a Tensorboard Summary to a dictionary.

    Accepts a tensorflow.summary.Summary, one encoded as a string,
    or a list of such encoded as strings.
    """
    values = {}
    if hasattr(tf_summary_str_or_pb, "summary"):
        summary_pb = tf_summary_str_or_pb.summary
        values[namespaced_tag("global_step", namespace)] = tf_summary_str_or_pb.step
        values["_timestamp"] = tf_summary_str_or_pb.wall_time
    elif isinstance(tf_summary_str_or_pb, (str, bytes, bytearray)):
        summary_pb = Summary()
        summary_pb.ParseFromString(tf_summary_str_or_pb)
    elif hasattr(tf_summary_str_or_pb, "__iter__"):
        summary_pb = [Summary() for _ in range(len(tf_summary_str_or_pb))]
        for i, summary in enumerate(tf_summary_str_or_pb):
            summary_pb[i].ParseFromString(summary)
            if i > 0:
                summary_pb[0].MergeFrom(summary_pb[i])
        summary_pb = summary_pb[0]
    else:
        summary_pb = tf_summary_str_or_pb

    if not hasattr(summary_pb, "value") or len(summary_pb.value) == 0:
        # Ignore these, caller is responsible for handling None
        return None

    def encode_images(_img_strs: List[bytes], _value: Any) -> None:
        try:
            from PIL import Image
        except ImportError:
            wandb.termwarn(
                "Install pillow if you are logging images with Tensorboard. "
                "To install, run `pip install pillow`.",
                repeat=False,
            )
            return None

        if len(_img_strs) == 0:
            return None

        images: List[Union["wandb.Video", "wandb.Image"]] = []
        for _img_str in _img_strs:
            # Supports gifs from TensorboardX
            if _img_str.startswith(b"GIF"):
                images.append(wandb.Video(io.BytesIO(_img_str), format="gif"))
            else:
                images.append(wandb.Image(Image.open(io.BytesIO(_img_str))))
        tag_idx = _value.tag.rsplit("/", 1)
        if len(tag_idx) > 1 and tag_idx[1].isdigit():
            tag, idx = tag_idx
            values.setdefault(history_image_key(tag, namespace), []).extend(images)
        else:
            values[history_image_key(_value.tag, namespace)] = images

        return None

    for value in summary_pb.value:
        kind = value.WhichOneof("value")
        if kind in IGNORE_KINDS:
            continue
        if kind == "simple_value":
            values[namespaced_tag(value.tag, namespace)] = value.simple_value
        elif kind == "tensor":
            plugin_name = value.metadata.plugin_data.plugin_name
            if plugin_name == "scalars" or plugin_name == "":
                values[namespaced_tag(value.tag, namespace)] = make_ndarray(
                    value.tensor
                )
            elif plugin_name == "images":
                img_strs = value.tensor.string_val[2:]  # First two items are dims.
                encode_images(img_strs, value)
            elif plugin_name == "histograms":
                # https://github.com/tensorflow/tensorboard/blob/master/tensorboard/plugins/histogram/summary_v2.py#L15-L26
                ndarray = make_ndarray(value.tensor)
                if ndarray is None:
                    continue
                shape = ndarray.shape
                counts = []
                bins = []
                if shape[0] > 1:
                    bins.append(ndarray[0][0])  # Add the left most edge
                    for v in ndarray:
                        counts.append(v[2])
                        bins.append(v[1])  # Add the right most edges
                elif shape[0] == 1:
                    counts = [ndarray[0][2]]
                    bins = ndarray[0][:2]
                if len(counts) > 0:
                    try:
                        # TODO: we should just re-bin if there are too many buckets
                        values[namespaced_tag(value.tag, namespace)] = wandb.Histogram(
                            np_histogram=(counts, bins)
                        )
                    except ValueError:
                        wandb.termwarn(
                            'Not logging key "{}". '
                            "Histograms must have fewer than {} bins".format(
                                namespaced_tag(value.tag, namespace),
                                wandb.Histogram.MAX_LENGTH,
                            ),
                            repeat=False,
                        )
            elif plugin_name == "pr_curves":
                pr_curve_data = make_ndarray(value.tensor)
                if pr_curve_data is None:
                    continue
                precision = pr_curve_data[-2, :].tolist()
                recall = pr_curve_data[-1, :].tolist()
                # TODO: (kdg) implement spec for showing additional info in tool tips
                # true_pos = pr_curve_data[1,:]
                # false_pos = pr_curve_data[2,:]
                # true_neg = pr_curve_data[1,:]
                # false_neg = pr_curve_data[1,:]
                # threshold = [1.0 / n for n in range(len(true_pos), 0, -1)]
                # min of each in case tensorboard ever changes their pr_curve
                # to allow for different length outputs
                data = []
                for i in range(min(len(precision), len(recall))):
                    # drop additional threshold values if they exist
                    if precision[i] != 0 or recall[i] != 0:
                        data.append((recall[i], precision[i]))
                # sort data so custom chart looks the same as tb generated pr curve
                # ascending recall, descending precision for the same recall values
                data = sorted(data, key=lambda x: (x[0], -x[1]))
                data_table = wandb.Table(data=data, columns=["recall", "precision"])
                name = namespaced_tag(value.tag, namespace)

                values[name] = custom_chart(
                    "wandb/line/v0",
                    data_table,
                    {"x": "recall", "y": "precision"},
                    {"title": f"{name} Precision v. Recall"},
                )
        elif kind == "image":
            img_str = value.image.encoded_image_string
            encode_images([img_str], value)
        # Coming soon...
        # elif kind == "audio":
        #     audio = wandb.Audio(
        #         six.BytesIO(value.audio.encoded_audio_string),
        #         sample_rate=value.audio.sample_rate,
        #         content_type=value.audio.content_type,
        #     )
        elif kind == "histo":
            tag = namespaced_tag(value.tag, namespace)
            if len(value.histo.bucket_limit) >= 3:
                first = (
                    value.histo.bucket_limit[0]
                    + value.histo.bucket_limit[0]
                    - value.histo.bucket_limit[1]
                )
                last = (
                    value.histo.bucket_limit[-2]
                    + value.histo.bucket_limit[-2]
                    - value.histo.bucket_limit[-3]
                )
                np_histogram = (
                    list(value.histo.bucket),
                    [first] + value.histo.bucket_limit[:-1] + [last],
                )
                try:
                    # TODO: we should just re-bin if there are too many buckets
                    values[tag] = wandb.Histogram(np_histogram=np_histogram)
                except ValueError:
                    wandb.termwarn(
                        f"Not logging key {tag!r}. "
                        f"Histograms must have fewer than {wandb.Histogram.MAX_LENGTH} bins",
                        repeat=False,
                    )
            else:
                # TODO: is there a case where we can render this?
                wandb.termwarn(
                    f"Not logging key {tag!r}. Found a histogram with only 2 bins.",
                    repeat=False,
                )
        # TODO(jhr): figure out how to share this between userspace and internal process or dont
        # elif value.tag == "_hparams_/session_start_info":
        #     if wandb.util.get_module("tensorboard.plugins.hparams"):
        #         from tensorboard.plugins.hparams import plugin_data_pb2
        #
        #         plugin_data = plugin_data_pb2.HParamsPluginData()        #
        #         plugin_data.ParseFromString(value.metadata.plugin_data.content)
        #         for key, param in six.iteritems(plugin_data.session_start_info.hparams):
        #             if not wandb.run.config.get(key):
        #                 wandb.run.config[key] = (
        #                     param.number_value or param.string_value or param.bool_value
        #                 )
        #     else:
        #         wandb.termerror(
        #             "Received hparams tf.summary, but could not import "
        #             "the hparams plugin from tensorboard"
        #         )
    return values


def reset_state() -> None:
    """Internal method for resetting state, called by wandb.finish()."""
    global STEPS
    STEPS = {"": {"step": 0}, "global": {"step": 0, "last_log": None}}


def _log(
    tf_summary_str_or_pb: Any,
    history: Optional["TBHistory"] = None,
    step: int = 0,
    namespace: str = "",
    **kwargs: Any,
) -> None:
    """Logs a tfsummary to wandb.

    Can accept a tf summary string or parsed event.  Will use wandb.run.history unless a
    history object is passed.  Can optionally namespace events.  Results are committed
    when step increases for this namespace.

    NOTE: This assumes that events being passed in are in chronological order
    """
    global STEPS
    global RATE_LIMIT_SECONDS
    # To handle multiple global_steps, we keep track of them here instead
    # of the global log
    last_step = STEPS.get(namespace, {"step": 0})

    # Commit our existing data if this namespace increased its step
    commit = False
    if last_step["step"] < step:
        commit = True

    log_dict = tf_summary_to_dict(tf_summary_str_or_pb, namespace)
    if log_dict is None:
        # not an event, just return
        return

    # Pass timestamp to history for loading historic data
    timestamp = log_dict.get("_timestamp", time.time())
    # Store our initial timestamp
    if STEPS["global"]["last_log"] is None:
        STEPS["global"]["last_log"] = timestamp
    # Rollup events that share the same step across namespaces
    if commit and step == STEPS["global"]["step"]:
        commit = False
    # Always add the biggest global_step key for non-default namespaces
    if step > STEPS["global"]["step"]:
        STEPS["global"]["step"] = step
    if namespace != "":
        log_dict["global_step"] = STEPS["global"]["step"]

    # Keep internal step counter
    STEPS[namespace] = {"step": step}

    if commit:
        # Only commit our data if we're below the rate limit or don't have one
        if (
            RATE_LIMIT_SECONDS is None
            or timestamp - STEPS["global"]["last_log"] >= RATE_LIMIT_SECONDS
        ):
            if history is None:
                if wandb.run is not None:
                    wandb.run._log({})
            else:
                history.add({})

        STEPS["global"]["last_log"] = timestamp

    if history is None:
        if wandb.run is not None:
            wandb.run._log(log_dict, commit=False)
    else:
        history._row_update(log_dict)


def log(tf_summary_str_or_pb: Any, step: int = 0, namespace: str = "") -> None:
    if wandb.run is None:
        raise wandb.Error(
            "You must call `wandb.init()` before calling `wandb.tensorflow.log`"
        )

    with telemetry.context() as tel:
        tel.feature.tensorboard_log = True

    _log(tf_summary_str_or_pb, namespace=namespace, step=step)



"""Implementation of Elastic Container Registry class for wandb launch."""
import base64
import logging
from typing import Dict, Tuple

from wandb.sdk.launch.environment.aws_environment import AwsEnvironment
from wandb.sdk.launch.utils import LaunchError
from wandb.util import get_module

from .abstract import AbstractRegistry

botocore = get_module(
    "botocore",
    required="AWS environment requires botocore to be installed. Please install "
    "it with `pip install wandb[launch]`.",
)

_logger = logging.getLogger(__name__)


class ElasticContainerRegistry(AbstractRegistry):
    """Elastic Container Registry class.

    Attributes:
        repo_name (str): The name of the repository.
        environment (AwsEnvironment): The AWS environment.
        uri (str): The uri of the repository.
    """

    repo_name: str
    environment: AwsEnvironment
    uri: str

    def __init__(self, repo_name: str, environment: AwsEnvironment) -> None:
        """Initialize the Elastic Container Registry.

        Arguments:
            repo_name (str): The name of the repository.
            environment (AwsEnvironment): The AWS environment.

        Raises:
            LaunchError: If there is an error verifying the registry.
        """
        super().__init__()
        _logger.info(
            f"Initializing Elastic Container Registry with repotisory {repo_name}."
        )
        self.repo_name = repo_name
        self.environment = environment
        self.verify()

    @classmethod
    def from_config(  # type: ignore[override]
        cls,
        config: Dict,
        environment: AwsEnvironment,
        verify: bool = True,
    ) -> "ElasticContainerRegistry":
        """Create an Elastic Container Registry from a config.

        Arguments:
            config (dict): The config.
            environment (AwsEnvironment): The AWS environment.

        Returns:
            ElasticContainerRegistry: The Elastic Container Registry.
        """
        if config.get("type") != "ecr":
            raise LaunchError(
                f"Could not create ElasticContainerRegistry from config. Expected type 'ecr' "
                f"but got '{config.get('type')}'."
            )
        repository = config.get("repository")
        if not repository:
            raise LaunchError(
                "Could not create ElasticContainerRegistry from config. 'repository' is required."
            )
        return cls(repository, environment)

    def verify(self) -> None:
        """Verify that the registry is accessible and the configured repo exists.

        Raises:
            RegistryError: If there is an error verifying the registry.
        """
        _logger.debug("Verifying Elastic Container Registry.")
        try:
            session = self.environment.get_session()
            client = session.client("ecr")
            response = client.describe_repositories(repositoryNames=[self.repo_name])
            self.uri = response["repositories"][0]["repositoryUri"].split("/")[0]

        except botocore.exceptions.ClientError as e:
            code = e.response["Error"]["Code"]
            msg = e.response["Error"]["Message"]
            # TODO: Log the code and the message here?
            raise LaunchError(
                f"Error verifying Elastic Container Registry: {code} {msg}"
            )

    def get_username_password(self) -> Tuple[str, str]:
        """Get the username and password for the registry.

        Returns:
            (str, str): The username and password.

        Raises:
            RegistryError: If there is an error getting the username and password.
        """
        _logger.debug("Getting username and password for Elastic Container Registry.")
        try:
            session = self.environment.get_session()
            client = session.client("ecr")
            response = client.get_authorization_token()
            username, password = base64.standard_b64decode(
                response["authorizationData"][0]["authorizationToken"]
            ).split(b":")
            return username.decode("utf-8"), password.decode("utf-8")

        except botocore.exceptions.ClientError as e:
            code = e.response["Error"]["Code"]
            msg = e.response["Error"]["Message"]
            # TODO: Log the code and the message here?
            raise LaunchError(f"Error getting username and password: {code} {msg}")

    def get_repo_uri(self) -> str:
        """Get the uri of the repository.

        Returns:
            str: The uri of the repository.
        """
        return self.uri + "/" + self.repo_name

    def check_image_exists(self, image_uri: str) -> bool:
        """Check if the image tag exists.

        Arguments:
            image_uri (str): The full image_uri.

        Returns:
            bool: True if the image tag exists.
        """
        uri, tag = image_uri.split(":")
        if uri != self.get_repo_uri():
            raise LaunchError(
                f"Image uri {image_uri} does not match Elastic Container Registry uri {self.get_repo_uri()}."
            )

        _logger.debug("Checking if image tag exists.")
        try:
            session = self.environment.get_session()
            client = session.client("ecr")
            response = client.describe_images(
                repositoryName=self.repo_name, imageIds=[{"imageTag": tag}]
            )
            return len(response["imageDetails"]) > 0

        except botocore.exceptions.ClientError as e:
            code = e.response["Error"]["Code"]
            if code == "ImageNotFoundException":
                return False
            msg = e.response["Error"]["Message"]
            raise LaunchError(f"Error checking if image tag exists: {code} {msg}")



import numbers
import os
from typing import TYPE_CHECKING, Optional, Type, Union

import wandb
from wandb import util
from wandb.sdk.lib import runid

from .._private import MEDIA_TMP
from ..base_types.media import Media

if TYPE_CHECKING:  # pragma: no cover
    from wandb.apis.public import Artifact as PublicArtifact

    from ...wandb_artifacts import Artifact as LocalArtifact
    from ...wandb_run import Run as LocalRun


class ImageMask(Media):
    """Format image masks or overlays for logging to W&B.

    Arguments:
        val: (dictionary)
            One of these two keys to represent the image:
                mask_data : (2D numpy array) The mask containing an integer class label
                    for each pixel in the image
                path : (string) The path to a saved image file of the mask
            class_labels : (dictionary of integers to strings, optional) A mapping of the
                integer class labels in the mask to readable class names. These will default
                to class_0, class_1, class_2, etc.

        key: (string)
            The readable name or id for this mask type (e.g. predictions, ground_truth)

    Examples:
        ### Logging a single masked image
        <!--yeadoc-test:log-image-mask-->
        ```python
        import numpy as np
        import wandb

        wandb.init()
        image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
        predicted_mask = np.empty((100, 100), dtype=np.uint8)
        ground_truth_mask = np.empty((100, 100), dtype=np.uint8)

        predicted_mask[:50, :50] = 0
        predicted_mask[50:, :50] = 1
        predicted_mask[:50, 50:] = 2
        predicted_mask[50:, 50:] = 3

        ground_truth_mask[:25, :25] = 0
        ground_truth_mask[25:, :25] = 1
        ground_truth_mask[:25, 25:] = 2
        ground_truth_mask[25:, 25:] = 3

        class_labels = {0: "person", 1: "tree", 2: "car", 3: "road"}

        masked_image = wandb.Image(
            image,
            masks={
                "predictions": {"mask_data": predicted_mask, "class_labels": class_labels},
                "ground_truth": {"mask_data": ground_truth_mask, "class_labels": class_labels},
            },
        )
        wandb.log({"img_with_masks": masked_image})
        ```

        ### Log a masked image inside a Table
        <!--yeadoc-test:log-image-mask-table-->
        ```python
        import numpy as np
        import wandb

        wandb.init()
        image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
        predicted_mask = np.empty((100, 100), dtype=np.uint8)
        ground_truth_mask = np.empty((100, 100), dtype=np.uint8)

        predicted_mask[:50, :50] = 0
        predicted_mask[50:, :50] = 1
        predicted_mask[:50, 50:] = 2
        predicted_mask[50:, 50:] = 3

        ground_truth_mask[:25, :25] = 0
        ground_truth_mask[25:, :25] = 1
        ground_truth_mask[:25, 25:] = 2
        ground_truth_mask[25:, 25:] = 3

        class_labels = {0: "person", 1: "tree", 2: "car", 3: "road"}

        class_set = wandb.Classes(
            [
                {"name": "person", "id": 0},
                {"name": "tree", "id": 1},
                {"name": "car", "id": 2},
                {"name": "road", "id": 3},
            ]
        )

        masked_image = wandb.Image(
            image,
            masks={
                "predictions": {"mask_data": predicted_mask, "class_labels": class_labels},
                "ground_truth": {"mask_data": ground_truth_mask, "class_labels": class_labels},
            },
            classes=class_set,
        )

        table = wandb.Table(columns=["image"])
        table.add_data(masked_image)
        wandb.log({"random_field": table})
        ```
    """

    _log_type = "mask"

    def __init__(self, val: dict, key: str) -> None:
        """Initialize an ImageMask object.

        Arguments:
            val: (dictionary) One of these two keys to represent the image:
                mask_data : (2D numpy array) The mask containing an integer class label
                    for each pixel in the image
                path : (string) The path to a saved image file of the mask
                class_labels : (dictionary of integers to strings, optional) A mapping
                    of the integer class labels in the mask to readable class names.
                    These will default to class_0, class_1, class_2, etc.

        key: (string)
            The readable name or id for this mask type (e.g. predictions, ground_truth)
        """
        super().__init__()

        if "path" in val:
            self._set_file(val["path"])
        else:
            np = util.get_module("numpy", required="Image mask support requires numpy")
            # Add default class mapping
            if "class_labels" not in val:
                classes = np.unique(val["mask_data"]).astype(np.int32).tolist()
                class_labels = {c: "class_" + str(c) for c in classes}
                val["class_labels"] = class_labels

            self.validate(val)
            self._val = val
            self._key = key

            ext = "." + self.type_name() + ".png"
            tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ext)

            pil_image = util.get_module(
                "PIL.Image",
                required='wandb.Image needs the PIL package. To get it, run "pip install pillow".',
            )
            image = pil_image.fromarray(val["mask_data"].astype(np.int8), mode="L")

            image.save(tmp_path, transparency=None)
            self._set_file(tmp_path, is_tmp=True, extension=ext)

    def bind_to_run(
        self,
        run: "LocalRun",
        key: Union[int, str],
        step: Union[int, str],
        id_: Optional[Union[int, str]] = None,
        ignore_copy_err: Optional[bool] = None,
    ) -> None:
        # bind_to_run key argument is the Image parent key
        # the self._key value is the mask's sub key
        super().bind_to_run(run, key, step, id_=id_, ignore_copy_err=ignore_copy_err)
        if hasattr(self, "_val") and "class_labels" in self._val:
            class_labels = self._val["class_labels"]

            run._add_singleton(
                "mask/class_labels",
                str(key) + "_wandb_delimeter_" + self._key,
                class_labels,
            )

    @classmethod
    def get_media_subdir(cls: Type["ImageMask"]) -> str:
        return os.path.join("media", "images", cls.type_name())

    @classmethod
    def from_json(
        cls: Type["ImageMask"], json_obj: dict, source_artifact: "PublicArtifact"
    ) -> "ImageMask":
        return cls(
            {"path": source_artifact.get_path(json_obj["path"]).download()},
            key="",
        )

    def to_json(self, run_or_artifact: Union["LocalRun", "LocalArtifact"]) -> dict:
        json_dict = super().to_json(run_or_artifact)

        if isinstance(run_or_artifact, wandb.wandb_sdk.wandb_run.Run):
            json_dict["_type"] = self.type_name()
            return json_dict
        elif isinstance(run_or_artifact, wandb.wandb_sdk.wandb_artifacts.Artifact):
            # Nothing special to add (used to add "digest", but no longer used.)
            return json_dict
        else:
            raise ValueError("to_json accepts wandb_run.Run or wandb_artifact.Artifact")

    @classmethod
    def type_name(cls: Type["ImageMask"]) -> str:
        return cls._log_type

    def validate(self, val: dict) -> bool:
        np = util.get_module("numpy", required="Image mask support requires numpy")
        # 2D Make this work with all tensor(like) types
        if "mask_data" not in val:
            raise TypeError(
                'Missing key "mask_data": An image mask requires mask data: a 2D array representing the predictions'
            )
        else:
            error_str = "mask_data must be a 2D array"
            shape = val["mask_data"].shape
            if len(shape) != 2:
                raise TypeError(error_str)
            if not (
                (val["mask_data"] >= 0).all() and (val["mask_data"] <= 255).all()
            ) and issubclass(val["mask_data"].dtype.type, np.integer):
                raise TypeError("Mask data must be integers between 0 and 255")

        # Optional argument
        if "class_labels" in val:
            for k, v in list(val["class_labels"].items()):
                if (not isinstance(k, numbers.Number)) or (not isinstance(v, str)):
                    raise TypeError(
                        "Class labels must be a dictionary of numbers to strings"
                    )
        return True



import inspect
import json
import re
import urllib
from copy import deepcopy
from typing import List as LList

from ... import __version__ as wandb_ver
from ... import termlog, termwarn
from ...sdk.lib import ipython
from ..public import Api as PublicApi
from ..public import RetryingClient
from ._blocks import P, PanelGrid, UnknownBlock, WeaveBlock, block_mapping, weave_blocks
from .mutations import UPSERT_VIEW, VIEW_REPORT
from .runset import Runset
from .util import Attr, Base, Block, coalesce, generate_name, nested_get, nested_set
from .validators import OneOf, TypeValidator


class Report(Base):
    project: str = Attr(json_path="viewspec.project.name")
    entity: str = Attr(json_path="viewspec.project.entityName")
    title: str = Attr(json_path="viewspec.displayName")
    description: str = Attr(json_path="viewspec.description")
    width: str = Attr(
        json_path="viewspec.spec.width",
        validators=[OneOf(["readable", "fixed", "fluid"])],
    )
    blocks: list = Attr(
        json_path="viewspec.spec.blocks",
        validators=[TypeValidator(Block, how="keys")],
    )

    def __init__(
        self,
        project,
        entity=None,
        title="Untitled Report",
        description="",
        width="readable",
        blocks=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._viewspec = self._default_viewspec()
        self._orig_viewspec = deepcopy(self._viewspec)

        self.project = project
        self.entity = coalesce(entity, PublicApi().default_entity, "")
        self.title = title
        self.description = description
        self.width = width
        self.blocks = coalesce(blocks, [])

    @classmethod
    def from_url(cls, url):
        report_id = cls._url_to_report_id(url)
        r = PublicApi().client.execute(
            VIEW_REPORT, variable_values={"reportId": report_id}
        )
        viewspec = r["view"]
        viewspec["spec"] = json.loads(viewspec["spec"])
        return cls.from_json(viewspec)

    @staticmethod
    def _url_to_report_id(url):
        try:
            report, *_ = url.split("?")
            # If the report title ends in trailing space
            report = report.replace("---", "--")
            *_, report_id = report.split("--")
        except ValueError as e:
            raise ValueError("Path must be `entity/project/reports/report_id`") from e
        else:
            return report_id

    @blocks.getter
    def blocks(self):
        json_path = self._get_path("blocks")
        block_specs = nested_get(self, json_path)
        blocks = []
        for bspec in block_specs:
            cls = block_mapping.get(bspec["type"], UnknownBlock)
            if cls is UnknownBlock:
                termwarn(
                    inspect.cleandoc(
                        f"""
                        UNKNOWN BLOCK DETECTED
                            This can happen if we have added new blocks, but you are using an older version of the SDK.
                            If your report is loading normally, you can safely ignore this message (but we recommend not touching UnknownBlock)
                            If you think this is an error, please file a bug report including your SDK version ({wandb_ver}) and this spec ({bspec})
                        """
                    )
                )
            if cls is WeaveBlock:
                for cls in weave_blocks:
                    try:
                        cls.from_json(bspec)
                    except Exception:
                        pass
                    else:
                        break
            blocks.append(cls.from_json(bspec))
        return blocks[1:-1]  # accounts for hidden p blocks

    @blocks.setter
    def blocks(self, new_blocks):
        json_path = self._get_path("blocks")
        new_block_specs = (
            [P("").spec] + [b.spec for b in new_blocks] + [P("").spec]
        )  # hidden p blocks
        nested_set(self, json_path, new_block_specs)

    @staticmethod
    def _default_viewspec():
        return {
            "id": None,
            "name": None,
            "spec": {
                "version": 5,
                "panelSettings": {},
                "blocks": [],
                "width": "readable",
                "authors": [],
                "discussionThreads": [],
                "ref": {},
            },
        }

    @classmethod
    def from_json(cls, viewspec):
        obj = cls(project=viewspec["project"]["name"])
        obj._viewspec = viewspec
        obj._orig_viewspec = deepcopy(obj._viewspec)
        return obj

    @property
    def viewspec(self):
        return self._viewspec

    @property
    def modified(self) -> bool:
        return self._viewspec != self._orig_viewspec

    @property
    def spec(self) -> dict:
        return self._viewspec["spec"]

    @property
    def client(self) -> "RetryingClient":
        return PublicApi().client

    @property
    def id(self) -> str:
        return self._viewspec["id"]

    @property
    def name(self) -> str:
        return self._viewspec["name"]

    @property
    def panel_grids(self) -> "LList[PanelGrid]":
        return [b for b in self.blocks if isinstance(b, PanelGrid)]

    @property
    def runsets(self) -> "LList[Runset]":
        return [rs for pg in self.panel_grids for rs in pg.runsets]

    @property
    def url(self) -> str:
        title = re.sub(r"\W", "-", self.title)
        title = re.sub(r"-+", "-", title)
        title = urllib.parse.quote(title)
        id = self.id.replace("=", "")
        app_url = PublicApi().client.app_url
        if not app_url.endswith("/"):
            app_url = app_url + "/"
        return f"{app_url}{self.entity}/{self.project}/reports/{title}--{id}"

    def save(self, draft: bool = False, clone: bool = False) -> "Report":
        if not self.modified:
            termwarn("Report has not been modified")

        # create project if not exists
        PublicApi().create_project(self.project, self.entity)

        # All panel grids must have at least one runset
        for pg in self.panel_grids:
            if not pg.runsets:
                pg.runsets = PanelGrid._default_runsets()

        # Check runsets with `None` for project and replace with the report's project.
        # We have to do this here because RunSets don't know about their report until they're added to it.
        for rs in self.runsets:
            rs.entity = coalesce(rs.entity, PublicApi().default_entity)
            rs.project = coalesce(rs.project, self.project)

        r = PublicApi().client.execute(
            UPSERT_VIEW,
            variable_values={
                "id": None if clone or not self.id else self.id,
                "name": generate_name() if clone or not self.name else self.name,
                "entityName": self.entity,
                "projectName": self.project,
                "description": self.description,
                "displayName": self.title,
                "type": "runs/draft" if draft else "runs",
                "spec": json.dumps(self.spec),
            },
        )

        viewspec = r["upsertView"]["view"]
        viewspec["spec"] = json.loads(viewspec["spec"])
        if clone:
            return Report.from_json(viewspec)
        else:
            self._viewspec["id"] = viewspec["id"]
            self._viewspec["name"] = viewspec["name"]
            return self

    def to_html(self, height: int = 1024, hidden: bool = False) -> str:
        """Generate HTML containing an iframe displaying this report."""
        try:
            url = self.url + "?jupyter=true"
            style = f"border:none;width:100%;height:{height}px;"
            prefix = ""
            if hidden:
                style += "display:none;"
                prefix = ipython.toggle_button("report")
            return prefix + f"<iframe src={url!r} style={style!r}></iframe>"
        except AttributeError:
            termlog("HTML repr will be available after you save the report!")

    def _repr_html_(self) -> str:
        return self.to_html()



"""This module contains the `WBTraceTree` media type, and the supporting dataclasses.

A `WBTraceTree` is a media object containing a root span and an
arbitrary model dump as a serializable dictionary. Logging such media type will
result in a W&B Trace Debugger panel being created in the workspace UI.
"""

import dataclasses
import hashlib
import json
import typing
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from wandb.data_types import _json_helper
from wandb.sdk.data_types import _dtypes
from wandb.sdk.data_types.base_types.media import Media

if TYPE_CHECKING:  # pragma: no cover
    from ..wandb_artifacts import Artifact as LocalArtifact
    from ..wandb_run import Run as LocalRun


class StatusCode(str, Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)


class SpanKind(str, Enum):
    LLM = "LLM"
    CHAIN = "CHAIN"
    AGENT = "AGENT"
    TOOL = "TOOL"

    def __str__(self) -> str:
        return str(self.value)


@dataclass()
class Result:
    inputs: Optional[Dict[str, Any]] = field(default=None)
    outputs: Optional[Dict[str, Any]] = field(default=None)


@dataclass()
class Span:
    span_id: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)
    start_time_ms: Optional[int] = field(default=None)
    end_time_ms: Optional[int] = field(default=None)
    status_code: Optional[StatusCode] = field(default=None)
    status_message: Optional[str] = field(default=None)
    attributes: Optional[Dict[str, Any]] = field(default=None)
    results: Optional[List[Result]] = field(default=None)
    child_spans: Optional[List["Span"]] = field(default=None)
    span_kind: Optional[SpanKind] = field(default=None)

    def add_attribute(self, key: str, value: Any) -> None:
        if self.attributes is None:
            self.attributes = {}
        self.attributes[key] = value

    def add_named_result(self, inputs: Dict[str, Any], outputs: Dict[str, Any]) -> None:
        if self.results is None:
            self.results = []
        self.results.append(Result(inputs, outputs))

    def add_child_span(self, span: "Span") -> None:
        if self.child_spans is None:
            self.child_spans = []
        self.child_spans.append(span)


class WBTraceTree(Media):
    """Media object for trace tree data.

    Arguments:
        root_span (Span): The root span of the trace tree.
        model_dict (dict, optional): A dictionary containing the model dump.
            NOTE: model_dict is a completely-user-defined dict. The UI will render
            a JSON viewer for this dict, giving special treatment to dictionaries
            with a `_kind` key. This is because model vendors have such different
            serialization formats that we need to be flexible here.
    """

    _log_type = "wb_trace_tree"

    def __init__(
        self,
        root_span: Span,
        model_dict: typing.Optional[dict] = None,
    ):
        super().__init__()
        self._root_span = root_span
        self._model_dict = model_dict

    @classmethod
    def get_media_subdir(cls) -> str:
        return "media/wb_trace_tree"

    def to_json(self, run: Optional[Union["LocalRun", "LocalArtifact"]]) -> dict:
        res = {}
        res["_type"] = self._log_type
        # Here we use `dumps` to put things into string format. This is because
        # the complex data structures create problems for gorilla history to parquet.
        if self._model_dict is not None:
            model_dump_str = _safe_serialize(self._model_dict)
            res["model_hash"] = _hash_id(model_dump_str)
            res["model_dict_dumps"] = model_dump_str
        res["root_span_dumps"] = _safe_serialize(dataclasses.asdict(self._root_span))
        return res

    def is_bound(self) -> bool:
        return True


class _WBTraceTreeFileType(_dtypes.Type):
    name = "wb_trace_tree"
    types = [WBTraceTree]


_dtypes.TypeRegistry.add(_WBTraceTreeFileType)


# generate a deterministic 16 character id based on input string
def _hash_id(s: str) -> str:
    return hashlib.md5(s.encode("utf-8")).hexdigest()[:16]


def _fallback_serialize(obj: Any) -> str:
    try:
        return f"<<non-serializable: {type(obj).__qualname__}>>"
    except Exception:
        return "<<non-serializable>>"


def _safe_serialize(obj: dict) -> str:
    try:
        return json.dumps(
            _json_helper(obj, None),
            skipkeys=True,
            default=_fallback_serialize,
        )
    except Exception:
        return "{}"



import threading
from collections import deque
from typing import TYPE_CHECKING, List

try:
    import psutil
except ImportError:
    psutil = None

from .aggregators import aggregate_mean
from .asset_registry import asset_registry
from .interfaces import Interface, Metric, MetricsMonitor

if TYPE_CHECKING:
    from typing import Deque

    from wandb.sdk.internal.settings_static import SettingsStatic


class NetworkSent:
    """Network bytes sent."""

    name = "network.sent"
    samples: "Deque[float]"

    def __init__(self) -> None:
        self.samples = deque([])
        self.sent_init = psutil.net_io_counters().bytes_sent

    def sample(self) -> None:
        self.samples.append(psutil.net_io_counters().bytes_sent - self.sent_init)

    def clear(self) -> None:
        self.samples.clear()

    def aggregate(self) -> dict:
        if not self.samples:
            return {}
        aggregate = aggregate_mean(self.samples)
        # todo: this is an adapter for the legacy metrics system
        # return {"network": {self.name: aggregate}}
        return {self.name: aggregate}


class NetworkRecv:
    """Network bytes received."""

    name = "network.recv"
    samples: "Deque[float]"

    def __init__(self) -> None:
        self.samples = deque([])
        self.recv_init = psutil.net_io_counters().bytes_recv

    def sample(self) -> None:
        self.samples.append(psutil.net_io_counters().bytes_recv - self.recv_init)

    def clear(self) -> None:
        self.samples.clear()

    def aggregate(self) -> dict:
        if not self.samples:
            return {}
        aggregate = aggregate_mean(self.samples)
        # todo: this is an adapter for the legacy metrics system
        # return {"network": {self.name: aggregate}}

        return {self.name: aggregate}


@asset_registry.register
class Network:
    def __init__(
        self,
        interface: "Interface",
        settings: "SettingsStatic",
        shutdown_event: threading.Event,
    ) -> None:
        self.name = self.__class__.__name__.lower()
        self.metrics: List[Metric] = [
            NetworkSent(),
            NetworkRecv(),
        ]
        self.metrics_monitor = MetricsMonitor(
            self.name,
            self.metrics,
            interface,
            settings,
            shutdown_event,
        )

    def start(self) -> None:
        self.metrics_monitor.start()

    def finish(self) -> None:
        self.metrics_monitor.finish()

    @classmethod
    def is_available(cls) -> bool:
        """Return a new instance of the CPU metrics."""
        return psutil is not None

    def probe(self) -> dict:
        """Return a dict of the hardware information."""
        # net_if_addrs = psutil.net_if_addrs()

        # return {
        #     self.name: {
        #         "interfaces": {
        #             k: {
        #                 "addresses": [
        #                     {
        #                         "address": v.address,
        #                         "netmask": v.netmask,
        #                         "broadcast": v.broadcast,
        #                         "ptp": v.ptp,
        #                     }
        #                     for v in v
        #                 ]
        #             }
        #             for k, v in net_if_addrs.items()
        #         }
        #     }
        # }
        return {}



import json
import os
import re
from typing import Any, Dict

from . import files as sm_files


def parse_sm_config() -> Dict[str, Any]:
    """Attempt to parse SageMaker configuration.

    Returns:
        A dictionary of SageMaker config keys/values or empty dict if not found.
    """
    conf = {}
    if os.path.exists(sm_files.SM_PARAM_CONFIG) and os.path.exists(
        sm_files.SM_RESOURCE_CONFIG
    ):
        conf["sagemaker_training_job_name"] = os.getenv("TRAINING_JOB_NAME")
        # Hyperparameter searches quote configs...
        for k, v in json.load(open(sm_files.SM_PARAM_CONFIG)).items():
            cast = v.strip('"')
            if re.match(r"^[-\d]+$", cast):
                cast = int(cast)
            elif re.match(r"^[-.\d]+$", cast):
                cast = float(cast)
            conf[k] = cast
    return conf



from asyncio import coroutine
from pytest import mark
from time import sleep
from promise import Promise


@mark.asyncio
@coroutine
def test_await():
    yield from Promise.resolve(True)


@mark.asyncio
@coroutine
def test_await_time():
    def resolve_or_reject(resolve, reject):
        sleep(.1)
        resolve(True)

    p = Promise(resolve_or_reject)
    assert p.get() is True


@mark.asyncio
@coroutine
def test_promise_coroutine():
    @coroutine
    def my_coro():
        yield from Promise.resolve(True)

    promise = Promise.resolve(my_coro())
    assert isinstance(promise, Promise)



from pytest import raises
import time
from promise import Promise, promisify, is_thenable


def test_benchmark_promise_creation(benchmark):
    @benchmark
    def create_promise():  # unnecessary function call
        p = Promise()


def test_benchmark_promise_resolve(benchmark):
    def create_promise():
        return Promise.resolve(True)

    result = benchmark(create_promise).get()
    assert result == True


def test_benchmark_is_thenable_basic_type(benchmark):
    def create_promise():
        return is_thenable(True)

    result = benchmark(create_promise)
    assert result == False


def test_benchmark_is_thenable_custom_type(benchmark):
    class MyType(object):
        pass

    my_type_instance = MyType()

    def create_promise():
        return is_thenable(my_type_instance)

    result = benchmark(create_promise)
    assert result == False


def test_benchmark_promise_creation_with_resolve(benchmark):
    do_resolve = lambda resolve, reject: resolve(True)

    def create_promise():  # unnecessary function call
        p = Promise(do_resolve)
        # p._wait()
        return p

    result = benchmark(create_promise).get()
    assert result == True


def test_benchmark_promise_creation_with_reject(benchmark):
    do_resolve = lambda resolve, reject: reject(Exception("Error"))

    def create_promise():  # unnecessary function call
        p = Promise(do_resolve)
        # p._wait()
        return p

    with raises(Exception) as exc_info:
        result = benchmark(create_promise).get()

    assert str(exc_info.value) == "Error"


# def test_benchmark_promisify_promise(benchmark):
#     instance = Promise()

#     def create_promise():  # unnecessary function call
#         return promisify(instance)

#     result = benchmark(create_promise)

#     assert isinstance(result, Promise)


def test_benchmark_promisify_custom_type(benchmark):
    class CustomThenable(object):
        pass
        # def then(self, resolve, reject):
        #     return resolve(True)

    instance = CustomThenable()

    def create_promise():  # unnecessary function call
        return Promise.resolve(instance)

    result = benchmark(create_promise)

    assert isinstance(result, Promise)
    assert result.get() == instance


def test_benchmark_promise_all(benchmark):
    values = range(1000)

    def create_promise():  # unnecessary function call
        return Promise.all(values)

    result = benchmark(create_promise)

    assert isinstance(result, Promise)
    assert result.get() == list(range(1000))


def test_benchmark_promise_all_promise(benchmark):
    values = [Promise.resolve(i) for i in range(100000)]

    def create_promise():  # unnecessary function call
        return Promise.all(values)

    result = benchmark(create_promise)

    assert isinstance(result, Promise)
    assert result.get() == list(range(100000))



import itertools

from ..language.ast import Document


def concat_ast(asts):
    return Document(definitions=list(itertools.chain.from_iterable(
        document.definitions for document in asts
    )))



import datetime
import logging
import shlex
import time
from typing import Any, Dict, Optional

if False:
    from google.cloud import aiplatform  # type: ignore   # noqa: F401

import yaml

import wandb
from wandb.apis.internal import Api
from wandb.util import get_module

from .._project_spec import LaunchProject, get_entry_point_command
from ..builder.abstract import AbstractBuilder
from ..builder.build import get_env_vars_dict
from ..environment.gcp_environment import GcpEnvironment
from ..utils import LOG_PREFIX, PROJECT_SYNCHRONOUS, LaunchError, run_shell
from .abstract import AbstractRun, AbstractRunner, Status

GCP_CONSOLE_URI = "https://console.cloud.google.com"

_logger = logging.getLogger(__name__)


class VertexSubmittedRun(AbstractRun):
    def __init__(self, job: Any) -> None:
        self._job = job

    @property
    def id(self) -> str:
        # numeric ID of the custom training job
        return self._job.name  # type: ignore

    @property
    def name(self) -> str:
        return self._job.display_name  # type: ignore

    @property
    def gcp_region(self) -> str:
        return self._job.location  # type: ignore

    @property
    def gcp_project(self) -> str:
        return self._job.project  # type: ignore

    def get_page_link(self) -> str:
        return "{console_uri}/vertex-ai/locations/{region}/training/{job_id}?project={project}".format(
            console_uri=GCP_CONSOLE_URI,
            region=self.gcp_region,
            job_id=self.id,
            project=self.gcp_project,
        )

    def wait(self) -> bool:
        self._job.wait()
        return self.get_status().state == "finished"

    def get_status(self) -> Status:
        job_state = str(self._job.state)  # extract from type PipelineState
        if job_state == "JobState.JOB_STATE_SUCCEEDED":
            return Status("finished")
        if job_state == "JobState.JOB_STATE_FAILED":
            return Status("failed")
        if job_state == "JobState.JOB_STATE_RUNNING":
            return Status("running")
        if job_state == "JobState.JOB_STATE_PENDING":
            return Status("starting")
        return Status("unknown")

    def cancel(self) -> None:
        self._job.cancel()


class VertexRunner(AbstractRunner):
    """Runner class, uses a project to create a VertexSubmittedRun."""

    def __init__(
        self, api: Api, backend_config: Dict[str, Any], environment: GcpEnvironment
    ) -> None:
        """Initialize a VertexRunner instance."""
        super().__init__(api, backend_config)
        self.environment = environment

    def run(
        self,
        launch_project: LaunchProject,
        builder: Optional[AbstractBuilder],
    ) -> Optional[AbstractRun]:
        """Run a Vertex job."""
        aiplatform = get_module(  # noqa: F811
            "google.cloud.aiplatform",
            "VertexRunner requires google.cloud.aiplatform to be installed",
        )
        resource_args = launch_project.resource_args.get("vertex")
        if not resource_args:
            resource_args = launch_project.resource_args.get("gcp-vertex")
        if not resource_args:
            raise LaunchError(
                "No Vertex resource args specified. Specify args via --resource-args with a JSON file or string under top-level key gcp_vertex"
            )
        gcp_staging_bucket = resource_args.get("staging_bucket")
        if not gcp_staging_bucket:
            raise LaunchError(
                "Vertex requires a staging bucket for training and dependency packages in the same region as compute. Specify a bucket under key staging_bucket."
            )
        gcp_machine_type = resource_args.get("machine_type") or "n1-standard-4"
        gcp_accelerator_type = (
            resource_args.get("accelerator_type") or "ACCELERATOR_TYPE_UNSPECIFIED"
        )
        gcp_accelerator_count = int(resource_args.get("accelerator_count") or 0)
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        gcp_training_job_name = resource_args.get(
            "job_name"
        ) or "{project}_{time}".format(
            project=launch_project.target_project, time=timestamp
        )
        service_account = resource_args.get("service_account")
        tensorboard = resource_args.get("tensorboard")
        aiplatform.init(
            project=self.environment.project,
            location=self.environment.region,
            staging_bucket=gcp_staging_bucket,
        )
        synchronous: bool = self.backend_config[PROJECT_SYNCHRONOUS]

        entry_point = launch_project.get_single_entry_point()

        if launch_project.docker_image:
            image_uri = launch_project.docker_image
        else:
            assert entry_point is not None
            assert builder is not None
            image_uri = builder.build_image(
                launch_project,
                entry_point,
            )

        # TODO: how to handle this?
        entry_cmd = get_entry_point_command(entry_point, launch_project.override_args)

        worker_pool_specs = [
            {
                "machine_spec": {
                    "machine_type": gcp_machine_type,
                    "accelerator_type": gcp_accelerator_type,
                    "accelerator_count": gcp_accelerator_count,
                },
                "replica_count": 1,
                "container_spec": {
                    "image_uri": image_uri,
                    "command": entry_cmd,
                    "env": [
                        {"name": k, "value": v}
                        for k, v in get_env_vars_dict(launch_project, self._api).items()
                    ],
                },
            }
        ]

        job = aiplatform.CustomJob(
            display_name=gcp_training_job_name, worker_pool_specs=worker_pool_specs
        )

        wandb.termlog(
            f"{LOG_PREFIX}Running training job {gcp_training_job_name} on {gcp_machine_type}."
        )

        if synchronous:
            job.run(service_account=service_account, tensorboard=tensorboard, sync=True)
        else:
            job.submit(
                service_account=service_account,
                tensorboard=tensorboard,
            )

        submitted_run = VertexSubmittedRun(job)

        while not getattr(job._gca_resource, "name", None):
            # give time for the gcp job object to be created and named, this should only loop a couple times max
            time.sleep(1)

        wandb.termlog(
            f"{LOG_PREFIX}View your job status and logs at {submitted_run.get_page_link()}."
        )
        return submitted_run


def get_gcp_config(config: str = "default") -> Any:
    return yaml.safe_load(
        run_shell(
            ["gcloud", "config", "configurations", "describe", shlex.quote(config)]
        )[0]
    )


def exists_on_gcp(image: str, tag: str) -> bool:
    out, err = run_shell(
        [
            "gcloud",
            "artifacts",
            "docker",
            "images",
            "list",
            shlex.quote(image),
            "--include-tags",
            f"--filter=tags:{shlex.quote(tag)}",
        ]
    )
    return tag in out and "sha256:" in out



import time
from typing import Any


class Timer:
    def __init__(self) -> None:
        self.start: float = time.perf_counter()
        self.stop: float = self.start

    def __enter__(self) -> "Timer":
        return self

    def __exit__(self, *args: Any) -> None:
        self.stop = time.perf_counter()

    @property
    def elapsed(self) -> float:
        return self.stop - self.start
